// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * Copyright (C) 2020 VeriSilicon Holdings Co., Ltd.
 *
 * This is vc8000e management driver for Linux.
 * vc8000e is a video encoder.
 */

#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/semaphore.h>
#include <linux/spinlock.h>
#include <linux/pci.h>
#include <linux/version.h>
#include <linux/vmalloc.h>
#include <linux/uaccess.h>

#include "common.h"
#include "vc8000e.h"
#include "transcoder.h"
#include "reg_dump.h"

#if (S1_VCE && S2_VCE)
#define MAX_SUBSYS 4
#else
#define MAX_SUBSYS 2
#endif

#if S1_VCE
#define SUBSYS_0_IO_ADDR (S1_VC8000E_OFF + 0x01000)
#define SUBSYS_0_IO_SIZE (40000 * 4) /* bytes */
#define S_ONEVF_VCE_0_IO_ADDR (ONE_VF_S1_VC8000E_OFF + 0x01000)

#define SUBSYS_1_IO_ADDR (S1_VC8000E_OFF + 0x05000)
#define SUBSYS_1_IO_SIZE (20000 * 4) /* bytes */
#define S_ONEVF_VCE_1_IO_ADDR (ONE_VF_S1_VC8000E_OFF + 0x05000)
#endif

#if S2_VCE
#define SUBSYS_2_IO_ADDR (S2_VC8000E_OFF + 0x01000)
#define SUBSYS_2_IO_SIZE (40000 * 4) /* bytes */
#define S_ONEVF_VCE_2_IO_ADDR (ONE_VF_S2_VC8000E_OFF + 0x01000)

#define SUBSYS_3_IO_ADDR (S2_VC8000E_OFF + 0x05000)
#define SUBSYS_3_IO_SIZE (20000 * 4) /* bytes */
#define S_ONEVF_VCE_3_IO_ADDR (ONE_VF_S2_VC8000E_OFF + 0x05000)
#endif
#define S_VCE_IO_ADDR  (VF_VC8000E_OFF + 0x05000)
#define S_VCE_IO_SIZE (20000 * 4)
#define S_VCE_0_IO_ADDR  (VF_VC8000E_OFF + 0x01000)
#define S_VCE_0_IO_SIZE (40000 * 4)

#define TIME_OUT_SEC 3

struct subsys_config {
	unsigned long base_addr;
	u32 iosize;
	u32 resouce_shared;
	u32 ctype;
	int slice_index;
};

struct vce_core_desc {
	u32 subsys_idx;
	u32 core_type;
	unsigned long offset;
	u32 reg_size;
	int irq;
	char *name;
	int slice_index;
};

struct subsys_data {
	struct subsys_config cfg;
	struct subsys_core_info core_info;
};

/*for all subsystem, the subsys info should be listed here for subsequent use*/
/*base_addr, iosize, resource_shared*/
static struct subsys_config pf_subsys_array[] = {
#if (S1_VCE && S2_VCE)
#if !SKIP_VCE
	{ SUBSYS_0_IO_ADDR, SUBSYS_0_IO_SIZE, 0, CORE_VC8000E, 0 }, //subsys_0
#endif
	{ SUBSYS_1_IO_ADDR, SUBSYS_1_IO_SIZE, 0, CORE_CUTREE, 0 }, //subsys_1
#if !SKIP_VCE
	{ SUBSYS_2_IO_ADDR, SUBSYS_2_IO_SIZE, 0, CORE_VC8000E, 1 }, //subsys_2
#endif
	{ SUBSYS_3_IO_ADDR, SUBSYS_3_IO_SIZE, 0, CORE_CUTREE, 1 } //subsys_3
#elif S1_VCE
#if !SKIP_VCE
	{ SUBSYS_0_IO_ADDR, SUBSYS_0_IO_SIZE, 0, CORE_VC8000E, 0 }, //subsys_0
#endif
	{ SUBSYS_1_IO_ADDR, SUBSYS_1_IO_SIZE, 0, CORE_CUTREE, 0 } //subsys_1
#elif S2_VCE
#if !SKIP_VCE
	{ SUBSYS_2_IO_ADDR, SUBSYS_2_IO_SIZE, 0, CORE_VC8000E, 1 }, //subsys_2
#endif
	{ SUBSYS_3_IO_ADDR, SUBSYS_3_IO_SIZE, 0, CORE_CUTREE, 1 } //subsys_3
#endif
};
static struct subsys_config vf1_subsys_array[] = {
#if !SKIP_VCE
	{ S_VCE_0_IO_ADDR, S_VCE_0_IO_SIZE, 0, CORE_VC8000E, 0 }, //subsys_0
#endif
	{ S_VCE_IO_ADDR, S_VCE_IO_SIZE, 0, CORE_CUTREE, 0 }
};
static struct subsys_config vf2_subsys_array[] = {
#if !SKIP_VCE
	{ S_VCE_0_IO_ADDR, S_VCE_0_IO_SIZE, 0, CORE_VC8000E, 1 }, //subsys_2
#endif
	{ S_VCE_IO_ADDR, S_VCE_IO_SIZE, 0, CORE_CUTREE, 1 }
};
static struct subsys_config vf_subsys_array[] = {
#if (S1_VCE && S2_VCE)
#if !SKIP_VCE
	{ S_ONEVF_VCE_0_IO_ADDR, SUBSYS_0_IO_SIZE, 0, CORE_VC8000E, 0 },
#endif
	{ S_ONEVF_VCE_1_IO_ADDR, SUBSYS_1_IO_SIZE, 0, CORE_CUTREE, 0 },
#if !SKIP_VCE
	{ S_ONEVF_VCE_2_IO_ADDR, SUBSYS_2_IO_SIZE, 0, CORE_VC8000E, 1 },
#endif
	{ S_ONEVF_VCE_3_IO_ADDR, SUBSYS_3_IO_SIZE, 0, CORE_CUTREE, 1 }
#elif S1_VCE
#if !SKIP_VCE
	{ S_ONEVF_VCE_0_IO_ADDR, SUBSYS_0_IO_SIZE, 0, CORE_VC8000E, 0 },
#endif
	{ S_ONEVF_VCE_1_IO_ADDR, SUBSYS_1_IO_SIZE, 0, CORE_CUTREE, 0 }
#elif S2_VCE
#if !SKIP_VCE
	{ S_ONEVF_VCE_2_IO_ADDR, SUBSYS_2_IO_SIZE, 0, CORE_VC8000E, 1 },
#endif
	{ S_ONEVF_VCE_3_IO_ADDR, SUBSYS_3_IO_SIZE, 0, CORE_CUTREE, 1 }
#endif
};
/*here config every core in all subsystem*/
/*NOTE: no matter what format(HEVC/H264/JPEG/AV1/...) is supported in VC8000E, just use [CORE_VC8000E] to indicate it's a VC8000E core*/
/*subsys_idx, core_type, offset, reg_size, irq*/
static struct vce_core_desc pf_core_array[] = {
#if (S1_VCE && S2_VCE)
#if !SKIP_VCE
	{ 0, CORE_VC8000E, 0x0, ENCODER_REGISTER_SIZE * 4, S1_VCE_IRQ,
	  "vc8000e_0", 0 }, //subsys_0_VC8000E
	{ 0, CORE_DEC400, 0x1000, DEC400_REGISTER_SIZE * 4, -1,
	  "null", 0 }, //subsys_0_DEC400
#endif
	{ 1, CORE_CUTREE, 0x0, IM_REGISTER_SIZE * 4, S1_IM_IRQ,
	  "im_0", 0 }, //subsys_1_CUTREE
#if !SKIP_VCE
	{ 2, CORE_VC8000E, 0x0, ENCODER_REGISTER_SIZE * 4, S2_VCE_IRQ,
	  "vc8000e_1", 1 }, //subsys_2_VC8000E
	{ 2, CORE_DEC400, 0x1000, DEC400_REGISTER_SIZE * 4, -1,
	  "null", 1 }, //subsys_2_DEC400
#endif
	{ 3, CORE_CUTREE, 0x0, IM_REGISTER_SIZE * 4, S2_IM_IRQ,
	  "im_1", 1 } //subsys_3_CUTREE
#elif S1_VCE
#if !SKIP_VCE
	{ 0, CORE_VC8000E, 0x0, ENCODER_REGISTER_SIZE * 4, S1_VCE_IRQ,
	  "vc8000e_0", 0 }, //subsys_0_VC8000E
	{ 0, CORE_DEC400, 0x1000, DEC400_REGISTER_SIZE * 4, -1,
	  "null", 0 }, //subsys_0_DEC400
#endif
	{ 1, CORE_CUTREE, 0x0, IM_REGISTER_SIZE * 4, S1_IM_IRQ,
	  "im_0", 0 } //subsys_1_CUTREE
#elif S2_VCE
#if !SKIP_VCE
	{ 0, CORE_VC8000E, 0x0, ENCODER_REGISTER_SIZE * 4, S2_VCE_IRQ,
	  "vc8000e_1", 1 }, //subsys_0_VC8000E
	{ 0, CORE_DEC400, 0x1000, DEC400_REGISTER_SIZE * 4, -1,
	  "null", 1 }, //subsys_0_DEC400
#endif
	{ 1, CORE_CUTREE, 0x0, IM_REGISTER_SIZE * 4, S2_IM_IRQ,
	  "im_1", 1 } //subsys_1_CUTREE
#endif
};
static struct vce_core_desc vf1_core_array[] = {
#if S1_VCE
#if !SKIP_VCE
	{ 0, CORE_VC8000E, 0x0, ENCODER_REGISTER_SIZE * 4, S1_VCE_IRQ,
	  "vc8000e_0", 0 }, //subsys_0_VC8000E
	{ 0, CORE_DEC400, 0x1000, DEC400_REGISTER_SIZE * 4, -1,
	  "null", 0 }, //subsys_0_DEC400
#endif
	{ 1, CORE_CUTREE, 0x0, IM_REGISTER_SIZE * 4, S1_IM_IRQ,
	  "im_0", 0 } //subsys_1_CUTREE
#endif
};
static struct vce_core_desc vf2_core_array[] = {
#if S2_VCE
#if !SKIP_VCE
	{ 0, CORE_VC8000E, 0x0, ENCODER_REGISTER_SIZE * 4, S2_VCE_IRQ,
	  "vc8000e_0", 1 }, //subsys_0_VC8000E
	{ 0, CORE_DEC400, 0x1000, DEC400_REGISTER_SIZE * 4, -1,
	  "null", 1 }, //subsys_0_DEC400
#endif
	{ 1, CORE_CUTREE, 0x0, IM_REGISTER_SIZE * 4, S2_IM_IRQ,
	  "im_0", 1 } //subsys_1_CUTREE
#endif
};

/* here's all the must remember stuff */
struct hantroenc_t {
	struct subsys_data
		subsys_data; //config of each core,such as base addr, iosize,etc
	u32 hw_id; //VC8000E/VC8000EJ hw id to indicate project
	u32 subsys_id; //subsys id for driver and sw internal use
	u32 is_valid; //indicate this subsys is hantro's core or not
	int pid[CORE_MAX]; //indicate which process is occupying the subsys
	u32 is_reserved[CORE_MAX]; //indicate this subsys is occupied by user or not
	u32 irq_received[CORE_MAX]; //indicate which core receives irq
	u32 irq_status[CORE_MAX]; //IRQ status of each core
	volatile u8 *hwregs;
	u32 job_id[CORE_MAX];
	unsigned long subsys_offset;
	struct enc_vce *tvce;
	struct file *filp[CORE_MAX];
};

struct enc_vce {
	int total_subsys_num;
	struct hantroenc_t hantroenc[MAX_SUBSYS];
	wait_queue_head_t reserve_queue;
	wait_queue_head_t enc_wait_queue;
	spinlock_t reserve_lock;
	spinlock_t irq_lock;
	struct sn_tranx_t *tdev;
	struct loading_info loading[MAX_SUBSYS];
	struct timer_list loading_timer;
	int cutree_used[2];    // IM_0, IM_1 status
	wait_queue_head_t reset_core_queue;
};

extern int total_vcmd_vc8000e_core_num;

static char *vce_get_coretype_str(int ct)
{
	switch (ct) {
		case (CORE_VC8000E):
			return ("VC8000E");
		case CORE_VC8000EJ:
			return ("VC8000EJ");
		case CORE_CUTREE:
			return ("CUTREE");
		case CORE_DEC400:
			return ("DEC400");
		case CORE_MMU:
			return ("MMU");
		case CORE_L2CACHE:
			return ("L2CACHE");
		case CORE_AXIFE:
			return ("AXIFE");
		case CORE_APBFT:
			return ("APBFT");
		case CORE_MMU_1:
			return ("MMU_1");
		case CORE_AXIFE_1:
			return ("AXIFE_1");
	default:
		return "Invalid core type";
	}
}

extern int hantrovcmd_init_ex(struct sn_tranx_t *tdev, int slice_index, u32 sub_module_type, u32 isIM);
extern u32 hantrovcmd_get_codec_core_num(struct sn_tranx_t *tdev, u32 module_type);

static int check_enc_irq(struct enc_vce *tvce, struct file *filp,
			 u32 *core_info, u32 *irq_status, u32 *job_id)
{
	unsigned long flags;
	int rdy = 0;
	u8 core_type = 0;
	u8 subsys_idx = 0;
	u32 loop = 30;

	core_type = (u8)(*core_info & 0x0F);
	subsys_idx = (u8)(*core_info >> 4);

	if (subsys_idx > tvce->total_subsys_num - 1) {
		*core_info = -1;
		*irq_status = 0;
		return 1;
	}

	do {
		spin_lock_irqsave(&tvce->irq_lock, flags);
		if ((tvce->hantroenc[subsys_idx].filp[core_type] != filp) ||
		    (tvce->hantroenc[subsys_idx].is_reserved[core_type] == 0)) {
			goto end_1;
		}

		if (tvce->hantroenc[subsys_idx].irq_received[core_type]) {
			rdy = 1;
			*core_info = subsys_idx;
			*irq_status = tvce->hantroenc[subsys_idx]
					      .irq_status[core_type];
			if (job_id != NULL)
				*job_id = tvce->hantroenc[subsys_idx]
						  .job_id[core_type];
			tvce->hantroenc[subsys_idx].irq_received[core_type] = 0;
			goto end_1;
		}

		spin_unlock_irqrestore(&tvce->irq_lock, flags);
		usleep_range(50, 100);
	} while (loop--);
	goto end_2;

end_1:
	spin_unlock_irqrestore(&tvce->irq_lock, flags);
end_2:
	return rdy;
}

static int check_enc_any_ready(struct enc_vce *tvce, struct file *filp,
			       struct core_wait_out *out)
{
	u32 i;
	int rdy = 1, ret;
	u32 core_info, irq_status, job_id;
	u32 core_type = CORE_VC8000E;

	for (i = 0; i < tvce->total_subsys_num; i++) {
		if (tvce->hantroenc[i].filp[core_type] == filp)
			if (!(tvce->hantroenc[i].subsys_data.core_info.type_info &
			      (1 << core_type)))
				continue;

		core_info = ((i << 4) | core_type);
		ret = check_enc_irq(tvce, filp, &core_info, &irq_status,
				    &job_id);
		if (ret && (core_info == i)) {
			out->job_id[out->irq_num] = job_id;
			out->irq_status[out->irq_num] = irq_status;
			out->irq_num++;
			rdy = 1;
		}
	}

	return rdy;
}

static int wait_enc_any_ready(struct enc_vce *tvce, struct file *filp,
			      struct core_wait_out *out)
{
	int ret;
	ret = wait_event_interruptible_timeout(
		tvce->enc_wait_queue, check_enc_any_ready(tvce, filp, out),
		TIME_OUT_SEC * HZ);
	if (!ret) {
		sn_pri(tvce->tdev, SN_ERR, "vce: wait %s timed out\n",
		       __FUNCTION__);
	} else if (ret > 0) {
		ret = 0;
	} else {
		sn_pri(tvce->tdev, SN_ERR,
		       "vce: wait_event_interruptible interrupted\n", ret);
	}
	return ret;

}

static int wait_enc_ready(struct enc_vce *tvce, struct file *filp,
			  u32 *core_info, u32 *irq_status)
{
	int ret;
	ret = wait_event_interruptible_timeout(
		tvce->enc_wait_queue,
		check_enc_irq(tvce, filp, core_info, irq_status, NULL),
		TIME_OUT_SEC * HZ);
	if (!ret) {
		sn_pri(tvce->tdev, SN_ERR, "vce: wait %s timed out\n",
		       __FUNCTION__);
	} else if (ret > 0) {
		ret = 0;
	} else {
		sn_pri(tvce->tdev, SN_ERR,
		       "vce: wait_event_interruptible interrupted\n", ret);
	}
	return ret;
}

static int check_core_occupation(struct hantroenc_t *dev, u8 core_type)
{
	int ret = 0;

	core_type = (core_type == CORE_VC8000EJ ? CORE_VC8000E : core_type);

	if (!dev->is_reserved[core_type]) {
		dev->is_reserved[core_type] = 1;
		dev->pid[core_type] = current->pid;
		ret = 1;
		sn_pri(dev->tvce->tdev, SN_DBG,
		       "vce: check_core_occupation pid=%d\n",
		       dev->pid[core_type]);
	}

	return ret;
}

static int get_idle_core(struct enc_vce *tvce, u32 *core_info,
			 u32 *core_info_tmp, struct file *filp)
{
	int ret = 0;
	u32 i = 0, cores, required_num = 0;
	u8 core_type = 0;
	static u32 reserved_job_id = 0;

	/*input core_info[32 bit]: mode[1bit](1:all 0:specified)+amount[3bit](the needing amount -1)+reserved+core_type[8bit]

	output core_info[32 bit]: the reserved core info to user space and defined as below.
	mode[1bit](1:all 0:specified)+amount[3bit](reserved total core num)+reserved+subsys_mapping[8bit]
	*/
	cores = *core_info;
	required_num = ((cores >> CORE_INFO_AMOUNT_OFFSET) & 0x7) + 1;
	core_type = (u8)(cores & 0xFF);

	if (*core_info_tmp == 0)
		*core_info_tmp = required_num << CORE_INFO_AMOUNT_OFFSET;
	else
		required_num = (*core_info_tmp >> CORE_INFO_AMOUNT_OFFSET);

	sn_pri(tvce->tdev, SN_DBG,
	       "vce: get_idle_core:required_num=%d,core_info=%x\n",
	       required_num, *core_info);

	if (required_num) {
		spin_lock(&tvce->reserve_lock);
		/* a valid free Core with specified core type */
		for (i = 0; i < tvce->total_subsys_num; i++) {
			if (tvce->hantroenc[i].subsys_data.core_info.type_info &
			    (1 << core_type)) {
				core_type = (core_type == CORE_VC8000EJ ?
						     CORE_VC8000E :
						     core_type);
				if (tvce->hantroenc[i].is_valid &&
				    check_core_occupation(&tvce->hantroenc[i],
							  core_type)) {
					*core_info_tmp =
						((((*core_info_tmp >>
						    CORE_INFO_AMOUNT_OFFSET) -
						   1)
						  << CORE_INFO_AMOUNT_OFFSET) |
						 (*core_info_tmp & 0x0FF));
					*core_info_tmp =
						(*core_info_tmp | (1 << i));
					if ((*core_info_tmp >>
					     CORE_INFO_AMOUNT_OFFSET) == 0) {
						ret = 1;
						*core_info =
							(reserved_job_id
							 << 16) |
							(*core_info_tmp & 0xFF);
						// tvce->hantroenc[i].job_id[core_type] = reserved_job_id;
						tvce->hantroenc[i]
							.job_id[core_type] =
							(reserved_job_id &
							 0xFFFF);
						reserved_job_id++;
						*core_info_tmp = 0;
						required_num = 0;
						if (tvce->hantroenc[i]
							    .filp[core_type] ==
						    NULL)
							tvce->hantroenc[i]
								.filp[core_type] =
								filp;
						break;
					}
				}
			}
		}

		if (required_num == 0) {
			if (core_type == CORE_CUTREE) {
				// only record IM used status
				if (i == 0 || i == 1) {
					tvce->cutree_used[0] = 1;
				} else if (i == 2 || i == 3) {
					tvce->cutree_used[1] = 1;
				}
			}
		}

		spin_unlock(&tvce->reserve_lock);
	} else
		ret = 1;

	sn_pri(tvce->tdev, SN_DBG, "vce: *core_info = %x\n", *core_info);
	return ret;
}

static int reserve_enc_core(struct enc_vce *tvce, struct file *filp,
			    u32 *core_info)
{
	u32 core_info_tmp = 0;
    u8 subsys_idx = (u8)((*core_info & 0xF0) >> 4);

	/* lock a core that has specified core id*/
	if (wait_event_interruptible(tvce->reserve_queue,
				     get_idle_core(tvce, core_info,
						   &core_info_tmp, filp) != 0))
		return -ERESTARTSYS;

	tvce->loading[subsys_idx].tv_s = ktime_get();

	return 0;
}

static void release_enc_core(struct enc_vce *tvce, struct file *filp,
			     u32 *core_info)
{
	u8 core_type = 0, subsys_idx = 0, pid = 0;
	pid = (u8)((*core_info) >> 31);

	subsys_idx = (u8)((*core_info & 0xF0) >> 4);
	core_type = (u8)(*core_info & 0x0F);

	sn_pri(tvce->tdev, SN_DBG,
	       "vce: release_enc_core:subsys_idx=%d,core_type=%x\n", subsys_idx,
	       core_type);
	/* release specified subsys and core type */

	if (tvce->hantroenc[subsys_idx].subsys_data.core_info.type_info &
	    (1 << core_type)) {
		core_type =
			(core_type == CORE_VC8000EJ ? CORE_VC8000E : core_type);
		spin_lock(&tvce->reserve_lock);
		sn_pri(tvce->tdev, SN_DBG,
		       "vce: subsys[%d].pid[%d]=%d,current->pid=%d\n",
		       subsys_idx, core_type,
		       tvce->hantroenc[subsys_idx].pid[core_type],
		       current->pid);
		if (tvce->hantroenc[subsys_idx].is_reserved[core_type] &&
		    (tvce->hantroenc[subsys_idx].pid[core_type] ==
			     current->pid ||
		     pid == 1)) {
			tvce->hantroenc[subsys_idx].pid[core_type] = -1;
			tvce->hantroenc[subsys_idx].is_reserved[core_type] = 0;
			tvce->hantroenc[subsys_idx].irq_received[core_type] = 0;
			tvce->hantroenc[subsys_idx].irq_status[core_type] = 0;
			tvce->hantroenc[subsys_idx].job_id[core_type] = 0;
			tvce->hantroenc[subsys_idx].filp[core_type] = NULL;
		} else {
			if (tvce->hantroenc[subsys_idx].pid[core_type] !=
				    current->pid &&
			    pid == 0)
				sn_pri(tvce->tdev, SN_DBG,
				       "vce: WARNING:pid(%d) is trying to release core reserved by pid(%d)\n",
				       current->pid,
				       tvce->hantroenc[subsys_idx]
					       .pid[core_type]);
		}
		if (core_type == CORE_CUTREE) {
			// release CUTREE CORE
			if (subsys_idx == 0 || subsys_idx == 1) {
				tvce->cutree_used[0] = 0;
			} else if (subsys_idx == 2 || subsys_idx == 3) {
				tvce->cutree_used[1] = 0;
			}
		}
		spin_unlock(&tvce->reserve_lock);
		if (core_type == CORE_CUTREE) {
			wake_up_interruptible_all(&tvce->reset_core_queue);
		}
		wake_up_interruptible_all(&tvce->reserve_queue);
	}

    tvce->loading[subsys_idx].tv_e = ktime_get();
    tvce->loading[subsys_idx].time_cnt +=
        ktime_to_us(ktime_sub(tvce->loading[subsys_idx].tv_e,
            tvce->loading[subsys_idx].tv_s));
}

static int vce_check_id(struct sn_tranx_t *tdev)
{
	int i;
	u32 hwid, found_hw = 0, hw_cfg;
	u32 val;

	struct enc_vce *tvce = tdev->modules[SN_MODULE_VC8000E];

	for (i = 0; i < tvce->total_subsys_num; i++) {
		tvce->hantroenc[i].hwregs =
			tdev->bar2_virt + tvce->hantroenc[i].subsys_offset;
		/*read hwid and check validness and store it*/
		hwid = readl((void *)tvce->hantroenc[i].hwregs);
		sn_pri(tdev, SN_DBG, "vce: hwid=0x%08x\n", hwid);

		/* check for encoder HW ID */
		if (((((hwid >> 16) & 0xFFFF) !=
		      ((ENC_HW_ID1 >> 16) & 0xFFFF))) &&
		    ((((hwid >> 16) & 0xFFFF) !=
		      ((ENC_HW_ID2 >> 16) & 0xFFFF)))) {
			sn_pri(tdev, SN_DBG, "vce: HW not found at 0x%lx\n",
			       tvce->hantroenc[i].subsys_data.cfg.base_addr);

			tvce->hantroenc[i].is_valid = 0;
			continue;
		}
		tvce->hantroenc[i].hw_id = hwid;
		tvce->hantroenc[i].is_valid = 1;
		found_hw = 1;

		hw_cfg = readl(tvce->hantroenc[i].hwregs + 320);
		tvce->hantroenc[i].subsys_data.core_info.type_info &=
			0xFFFFFFFC;
		if (hw_cfg & 0x88000000)
			tvce->hantroenc[i].subsys_data.core_info.type_info |=
				(1 << CORE_VC8000E);
		if (hw_cfg & 0x00008000)
			tvce->hantroenc[i].subsys_data.core_info.type_info |=
				(1 << CORE_VC8000EJ);

		/* disable dec400 interrupt */
		val = readl(tdev->bar2_virt +
			    (tvce->hantroenc[i].subsys_offset - 0x1000) + 0x64);
		writel(val | (1 << 2),
		       tdev->bar2_virt + (tvce->hantroenc[i].subsys_offset - 0x1000) +
			       0x64);

		sn_pri(tdev, SN_DBG,
		       "vce: HW at base <0x%lx> with ID <0x%08x>\n",
		       tvce->hantroenc[i].subsys_data.cfg.base_addr, hwid);
	}

	if (found_hw == 0) {
		sn_pri(tdev, SN_ERR, "vce: NO ANY HW found!!\n");
		return -1;
	}

	return 0;
}
void get_shadow_isr_off(struct hantroenc_t *dev, u32 *shadow_isr_off, u32 ctype)
{
	u8 pf_vf_mode = dev->tvce->tdev->pf_vf_mode;
	switch(pf_vf_mode) {
		case PF_MODE:
			if(ctype == CORE_VC8000E) {
#if (S1_VCE && S2_VCE)
				if (dev->subsys_id == 0)
					*shadow_isr_off = S1_INT_SD_VCE_ISR;
				else if (dev->subsys_id == 2)
					*shadow_isr_off = S2_INT_SD_VCE_ISR;
#elif S1_VCE
				*shadow_isr_off = S1_INT_SD_VCE_ISR;
#elif S2_VCE
				*shadow_isr_off = S2_INT_SD_VCE_ISR;
#endif
			}
			else if(ctype == CORE_CUTREE) {
#if (S1_VCE && S2_VCE)
				if (dev->subsys_id == 1)
					*shadow_isr_off = S1_INT_SD_VCE_LA_ISR;
				else if (dev->subsys_id == 3)
					*shadow_isr_off = S2_INT_SD_VCE_LA_ISR;
#elif S1_VCE
				*shadow_isr_off = S1_INT_SD_VCE_LA_ISR;
#elif S2_VCE
				*shadow_isr_off = S2_INT_SD_VCE_LA_ISR;
#endif
			}
			break;
		case TWO_VF_MODE:
			if(ctype == CORE_VC8000E) {
				*shadow_isr_off = VF_INT_SD_VCE_ISR;
			}
			else if(ctype == CORE_CUTREE) {
				*shadow_isr_off = VF_INT_SD_VCE_LA_ISR;
			}
			break;
		case ONE_VF_MODE:
			if(ctype == CORE_VC8000E) {
#if (S1_VCE && S2_VCE)
				if (dev->subsys_id == 0)
					*shadow_isr_off = ONE_VF_S1_INT_SD_VCE_ISR;
				else if (dev->subsys_id == 2)
					*shadow_isr_off = ONE_VF_S2_INT_SD_VCE_ISR;
#elif S1_VCE
				*shadow_isr_off = ONE_VF_S1_INT_SD_VCE_ISR;
#elif S2_VCE
				*shadow_isr_off = ONE_VF_S2_INT_SD_VCE_ISR;
#endif
			}
			else if(ctype == CORE_CUTREE) {
#if (S1_VCE && S2_VCE)
				if (dev->subsys_id == 1)
					*shadow_isr_off = ONE_VF_S1_INT_SD_VCE_LA_ISR;
				else if (dev->subsys_id == 3)
					*shadow_isr_off = ONE_VF_S2_INT_SD_VCE_LA_ISR;
#elif S1_VCE
				*shadow_isr_off = ONE_VF_S1_INT_SD_VCE_LA_ISR;
#elif S2_VCE
				*shadow_isr_off = ONE_VF_S2_INT_SD_VCE_LA_ISR;
#endif
			}
			break;
		default:
			sn_pri(dev->tvce->tdev, SN_INF, "%s,pf_vf_mode incorrect\n",
			       __func__);
			break;
	}
}
static irqreturn_t hantroenc_isr(int irq, void *dev_id)
{
	unsigned int handled = 0;
	struct hantroenc_t *dev = (struct hantroenc_t *)dev_id;
	struct enc_vce *tvce = dev->tvce;
	u32 irq_status, enable_status;
	unsigned long flags;
	u32 hwId, majorId, wClr;
	u32 ctype = 0;
	volatile u8 *hwregs = NULL;

	u32 int_shadow_isr_off = 0;

	spin_lock_irqsave(&tvce->irq_lock, flags);

	if (irq == dev->subsys_data.core_info.irq[CORE_VC8000E]) {
		ctype = CORE_VC8000E;
		get_shadow_isr_off(dev, &int_shadow_isr_off, ctype);
	}
	else if (irq == dev->subsys_data.core_info.irq[CORE_CUTREE]) {
		ctype = CORE_CUTREE;
		get_shadow_isr_off(dev, &int_shadow_isr_off, ctype);
	} else
		sn_pri(tvce->tdev, SN_ERR, "vce: irq:%d error\n", irq);

	hwregs = dev->hwregs + dev->subsys_data.core_info.offset[ctype];

	irq_status = readl(hwregs + 0x04);
	sn_pri(tvce->tdev, SN_DBG,
		"vce: irq:%d received! core:%d, vce status:0x%x\n",
		irq, ctype, irq_status);
	enable_status = readl(hwregs + 20);
	if (irq_status & ASIC_STATUS_ALL) {
		if (irq_status & 0x20)
			writel(0, hwregs + 0x14);

		/* clear all IRQ bits. (hwId >= 0x80006100) means IRQ is cleared by writting 1 */
		hwId = readl(hwregs);
		majorId = (hwId & 0x0000FF00) >> 8;
		wClr = (majorId >= 0x61) ? irq_status : (irq_status & (~0x1FD));
		writel(wClr, hwregs + 0x04);
		if ((readl(hwregs + 0x04)) & ASIC_STATUS_FRAME_READY)
			irq_status |= ASIC_STATUS_FRAME_READY;
		dev->irq_received[ctype] = 1;
		dev->irq_status[ctype] = irq_status;

		handled++;
	}
	spin_unlock_irqrestore(&tvce->irq_lock, flags);

	if (handled)
		wake_up_interruptible_all(&tvce->enc_wait_queue);

	return IRQ_HANDLED;
}


/*
 * Calculate the utilization of encoder in one second.
 * It is encoder used time, from reserve a idle
 * core to release it, add up the time period in one second.
 * So the total used time divide one second is the utilization.
 */
static void enc_loading_timer_isr(struct timer_list *t)
{
	struct enc_vce *tvce = from_timer(tvce, t, loading_timer);
	int i = 0;

	mod_timer(&tvce->loading_timer, jiffies + LOADING_TIME*HZ);

	for (i = 0; i < MAX_SUBSYS; i++) {
		tvce->loading[i].time_cnt_saved = tvce->loading[i].time_cnt;
		tvce->loading[i].time_cnt = 0;
		tvce->loading[i].total_time = LOADING_TIME * 1000000;
	}
}

/* Display encoder utilization which is the average of the two core. */
static ssize_t enc_util_show(struct device *dev,
				 struct device_attribute *attr,
				 char *buf)
{
	struct sn_misc_tdev *mtdev = dev_get_drvdata(dev);
	struct sn_tranx_t *tdev = mtdev->tdev;
	struct enc_vce *tvce = tdev->modules[SN_MODULE_VC8000E];
	unsigned long average_loading, saved = 0, total = 0;
	int i = 0;

	for (i = 0; i < MAX_SUBSYS; i++) {
		saved += tvce->loading[i].time_cnt_saved;
		total += tvce->loading[i].total_time;
	}
	average_loading = saved * 100 / total;
	if (average_loading > 100)
		average_loading = 100;

	return sprintf(buf, "%ld%%\n", average_loading);
}

static DEVICE_ATTR_RO(enc_util);

static struct attribute *trans_enc_sysfs_entries[] = {
	&dev_attr_enc_util.attr,
	NULL
};

static struct attribute_group trans_enc_attribute_group = {
	.name = NULL,
	.attrs = trans_enc_sysfs_entries,
};

static struct subsys_config* get_total_subsys_num(struct enc_vce *tvce,
								int* total_subsys_num)
{
	struct subsys_config *ret_subsys_pointer = NULL;
	u8 pf_vf_mode = tvce->tdev->pf_vf_mode;
	switch (pf_vf_mode) {
		case PF_MODE:
			if (tvce->tdev->vf_index == PF_INDEX) {
				*total_subsys_num =
				sizeof(pf_subsys_array) / sizeof(pf_subsys_array[0]);
				ret_subsys_pointer = pf_subsys_array;
			}
			else {
				sn_pri(tvce->tdev, SN_ERR, "%s,vf_index incorrect in PF_MODE\n",
						__func__);
			}
			break;
		case TWO_VF_MODE:
			if (tvce->tdev->vf_index == VF2_INDEX) {
				*total_subsys_num =
				sizeof(vf2_subsys_array) / sizeof(vf2_subsys_array[0]);
				ret_subsys_pointer = vf2_subsys_array;
			}
			else if (tvce->tdev->vf_index == VF1_INDEX) {
				*total_subsys_num =
				sizeof(vf1_subsys_array) / sizeof(vf1_subsys_array[0]);
				ret_subsys_pointer = vf1_subsys_array;
			}
			else {
				sn_pri(tvce->tdev, SN_ERR, "%s,vf_index incorrect in TWO_VF_MODE\n",
						__func__);
			}
			break;
		case ONE_VF_MODE:
			if (tvce->tdev->vf_index == VF1_INDEX) {
				*total_subsys_num =
				sizeof(vf_subsys_array) / sizeof(vf_subsys_array[0]);
				ret_subsys_pointer = vf_subsys_array;
			}
			else {
				sn_pri(tvce->tdev, SN_ERR, "%s,vf_index incorrect in ONE_VF_MODE\n",
						__func__);
			}
			break;
		default:
			sn_pri(tvce->tdev, SN_INF, "%s,vf_max_count incorrect\n",
					__func__);
			break;
	}
	return ret_subsys_pointer;
}
struct vce_core_desc* get_current_core_array(struct enc_vce *tvce,
						int *total_core_num)
{
	struct vce_core_desc* ret_core_pointer = NULL;
	u8 pf_vf_mode = tvce->tdev->pf_vf_mode;
	switch (pf_vf_mode) {
		case PF_MODE:
			if (tvce->tdev->vf_index == PF_INDEX) {
				*total_core_num =
				sizeof(pf_core_array) / sizeof(pf_core_array[0]);
				ret_core_pointer = pf_core_array;
			}
			else {
				sn_pri(tvce->tdev, SN_ERR, "%s,vf_index incorrect in PF_MODE\n",
						__func__);
			}
			break;
		case TWO_VF_MODE:
			if (tvce->tdev->vf_index == VF2_INDEX) {
				*total_core_num =
				sizeof(vf2_core_array) / sizeof(vf2_core_array[0]);
				ret_core_pointer = vf2_core_array;
			}
			else if (tvce->tdev->vf_index == VF1_INDEX) {
				*total_core_num =
				sizeof(vf1_core_array) / sizeof(vf1_core_array[0]);
				ret_core_pointer = vf1_core_array;
			}
			else {
				sn_pri(tvce->tdev, SN_ERR, "%s,vf_index incorrect in TWO_VF_MODE\n",
						__func__);
			}
			break;
		case ONE_VF_MODE:
			if (tvce->tdev->vf_index == VF1_INDEX) {
				*total_core_num =
				sizeof(pf_core_array) / sizeof(pf_core_array[0]);
				ret_core_pointer = pf_core_array;
			}
			else {
				sn_pri(tvce->tdev, SN_ERR, "%s,vf_index incorrect in ONE_VF_MODE\n",
						__func__);
			}
			break;
		default:
			sn_pri(tvce->tdev, SN_INF, "%s,vf_max_count incorrect\n",
					__func__);
			break;
	}
	return ret_core_pointer;
}
static void disable_vcmd_gate(struct enc_vce *tvce)
{
	u8 pf_vf_mode = tvce->tdev->pf_vf_mode;
	switch (pf_vf_mode) {
		case PF_MODE:
#if S2_VCE
#if !SKIP_VCE
			writel(0x0000FFFE, tvce->tdev->bar2_virt +
					S2_VC8000E_OFF + 0x64);
#endif
			writel(0x0000FFFD, tvce->tdev->bar2_virt +
					S2_VC8000E_OFF + 0x4000 + 0x64);
#endif
#if S1_VCE
#if !SKIP_VCE
			writel(0x0000FFFE, tvce->tdev->bar2_virt +
					S1_VC8000E_OFF + 0x64);
#endif
			writel(0x0000FFFD, tvce->tdev->bar2_virt +
					S1_VC8000E_OFF + 0x4000 + 0x64);
#endif
			break;
		case TWO_VF_MODE:
#if S2_VCE
#if !SKIP_VCE
			writel(0x0000FFFE, tvce->tdev->bar2_virt +
					VF_VC8000E_OFF + 0x64);
#endif
			writel(0x0000FFFD, tvce->tdev->bar2_virt +
					VF_VC8000E_OFF + 0x4000 + 0x64);
#endif
#if S1_VCE
#if !SKIP_VCE
			writel(0x0000FFFE, tvce->tdev->bar2_virt +
					VF_VC8000E_OFF + 0x64);
#endif
			writel(0x0000FFFD, tvce->tdev->bar2_virt +
					VF_VC8000E_OFF + 0x4000 + 0x64);
#endif
			break;
		case ONE_VF_MODE:
#if S2_VCE
#if !SKIP_VCE
			writel(0x0000FFFE, tvce->tdev->bar2_virt +
					ONE_VF_S2_VC8000E_OFF + 0x64);
#endif
			writel(0x0000FFFD, tvce->tdev->bar2_virt +
					ONE_VF_S2_VC8000E_OFF + 0x4000 + 0x64);
#endif
#if S1_VCE
#if !SKIP_VCE
			writel(0x0000FFFE, tvce->tdev->bar2_virt +
					ONE_VF_S1_VC8000E_OFF + 0x64);
#endif
			writel(0x0000FFFD, tvce->tdev->bar2_virt +
					ONE_VF_S1_VC8000E_OFF + 0x4000 + 0x64);
#endif
			break;
		default:
			sn_pri(tvce->tdev, SN_INF, "%s,vf_max_count incorrect\n",
			       __func__);
			break;
	}
}

static int vc8000e_subsys_reset_keep(struct sn_tranx_t *tdev)
{
#if S1_VCE
	sys_config_reset_keep(tdev, 0, SYS_CTL_VCE);
#endif

#if S2_VCE
	sys_config_reset_keep(tdev, 1, SYS_CTL_VCE);
#endif

	return 0;
}

static int vc8000e_subsys_reset_release(struct sn_tranx_t *tdev)
{
#if S1_VCE
	sys_config_reset_release(tdev, 0, SYS_CTL_VCE);
#endif

#if S2_VCE
	sys_config_reset_release(tdev, 1, SYS_CTL_VCE);
#endif


#if VCMD_ENABLE_VC8000E

#if S1_VCE
	hantrovcmd_init_ex(tdev, 0, SYS_CTL_VCE, 0);
	hantrovcmd_init_ex(tdev, 0, SYS_CTL_VCE, 1);
#endif

#if S2_VCE
	hantrovcmd_init_ex(tdev, 1, SYS_CTL_VCE, 0);
	hantrovcmd_init_ex(tdev, 1, SYS_CTL_VCE, 1);
#endif

#else

	switch (tdev->pf_vf_mode) {
		case PF_MODE:
#if S1_VCE && !SKIP_VCE
			writel(0xFFFFFFFE, tdev->bar2_virt + S1_VC8000E_OFF + 0x64);
			writel(0xFFFFFFFD, tdev->bar2_virt + S1_VC8000E_OFF + 0x4000 + 0x64);
#endif
#if S2_VCE && !SKIP_VCE
			writel(0xFFFFFFFE, tdev->bar2_virt + S2_VC8000E_OFF + 0x64);
			writel(0xFFFFFFFD, tdev->bar2_virt + S2_VC8000E_OFF + 0x4000 + 0x64);
#endif
	break;
		case ONE_VF_MODE:
#if S1_VCE && !SKIP_VCE
			writel(0xFFFFFFFE, tdev->bar2_virt + ONE_VF_S1_VC8000E_OFF + 0x64);
			writel(0xFFFFFFFD, tdev->bar2_virt + ONE_VF_S1_VC8000E_OFF + 0x4000 + 0x64);
#endif
#if S2_VCE && !SKIP_VCE
			writel(0xFFFFFFFE, tdev->bar2_virt + ONE_VF_S2_VC8000E_OFF + 0x64);
			writel(0xFFFFFFFD, tdev->bar2_virt + ONE_VF_S2_VC8000E_OFF + 0x4000 + 0x64);
#endif
		break;
		case TWO_VF_MODE:
#if S1_VCE && !SKIP_VCE
			writel(0xFFFFFFFE, tdev->bar2_virt + VF_VC8000E_OFF + 0x64);
			writel(0xFFFFFFFD, tdev->bar2_virt + VF_VC8000E_OFF + 0x4000 + 0x64);
#endif
#if S2_VCE && !SKIP_VCE
			writel(0xFFFFFFFE, tdev->bar2_virt + VF_VC8000E_OFF + 0x64);
			writel(0xFFFFFFFD, tdev->bar2_virt + VF_VC8000E_OFF + 0x4000 + 0x64);
#endif
		break;
	}
#endif

	return 0;
}

static int vc8000e_soft_reset(struct sn_tranx_t *tdev)
{
#if !VCMD_ENABLE_VC8000E
	int ret = 0, i = 0, j = 0;
	u32 val = 0;
	struct enc_vce *tvce = tdev->modules[SN_MODULE_VC8000E];
#endif
	vc8000e_subsys_reset_keep(tdev);
	vc8000e_subsys_reset_release(tdev);
#if !VCMD_ENABLE_VC8000E
	if (tvce == NULL) {
		sn_pri(tdev, SN_ERR, "vce: it's null please check!!!\n");
		return -EFAULT;
	}
	ret = vce_check_id(tdev);
	if (ret < 0) {
		for (i = 0; i < tvce->total_subsys_num; i++) {
			for (j = 0; j < i; j++) {
				val = tvce->hantroenc[i].subsys_data.core_info.irq[j];
				if (val > 0)
					free_irq(val, (void *)&tvce->hantroenc[i]);
			}
		}
		kfree(tvce);
		tdev->modules[SN_MODULE_VC8000E] = NULL;
		sn_pri(tdev, SN_ERR, "vce: module reset failed\n");
		return ret;
	}
#endif

	sn_pri(tdev, SN_INF, "vce: soft reset done.\n");
	return 0;

}

static void vce_reset_slice_dec400(struct sn_tranx_t *tdev, u32 id)
{
	struct enc_vce *tvce = tdev->modules[SN_MODULE_VC8000E];
	u32 val;

    /* disable dec400 interrupt */
	val = readl(tdev->bar2_virt +
			(tvce->hantroenc[id].subsys_offset - 0x1000) + 0x64);
	writel(val | (1 << 2),
			tdev->bar2_virt + (tvce->hantroenc[id].subsys_offset - 0x1000) +
			0x64);
}

static int vc8000e_soft_reset_slice(struct sn_tranx_t *tdev, u32 core_id)
{
	int slice = 0;
	struct enc_vce *tvce = tdev->modules[SN_MODULE_VC8000E];

	if (core_id == 0)
		slice = 0;
	else if (core_id == 2)
		slice = 1;
	else
		return 0;

	wait_event_interruptible(tvce->reset_core_queue, !(tvce->cutree_used[slice]));
	spin_lock(&tvce->reserve_lock);
	sys_config_reset_keep(tdev, slice, SYS_CTL_VCE);
	sys_config_reset_release(tdev, slice, SYS_CTL_VCE);
	vce_reset_slice_dec400(tdev, core_id);
	spin_unlock(&tvce->reserve_lock);
	wake_up_interruptible_all(&tvce->reserve_queue);

	sn_pri(tdev, SN_INF, "vce: soft reset slice %d done.\n", slice);
	return 0;

}

int vc8000e_init(struct sn_tranx_t *tdev)
{
	int ret = -1, i;
	u32 val, id, ctype;
	unsigned long bar2_base;
	struct enc_vce *tvce;
	int total_core_num;
	int j, m, n;
	struct subsys_config *subsys_array = NULL;
	struct vce_core_desc *core_array = NULL;

	tvce = (struct enc_vce *)kzalloc(sizeof(*tvce), GFP_KERNEL);
	if (!tvce)
		return -ENOMEM;
	memset(tvce, 0, sizeof(*tvce));

	tdev->modules[SN_MODULE_VC8000E] = tvce;
	tvce->tdev = tdev;

	subsys_array = get_total_subsys_num( tvce, &(tvce->total_subsys_num));
	if(subsys_array == NULL) {
		sn_pri(tvce->tdev, SN_ERR, "%s,subsys_array is NULL\n",
			       __func__);
		goto err;
	}

	bar2_base = pci_resource_start(tdev->pdev, 2);
	for (i = 0; i < tvce->total_subsys_num; i++) {
		tvce->hantroenc[i].subsys_data.cfg = *(subsys_array + i);
		tvce->hantroenc[i].subsys_data.cfg.base_addr += bar2_base;
		tvce->hantroenc[i].subsys_id = i;
		tvce->hantroenc[i].tvce = tvce;
		tvce->hantroenc[i].subsys_offset = (subsys_array + i)->base_addr;
	}
	tvce->cutree_used[0] = tvce->cutree_used[1] = 0;
	core_array = get_current_core_array(tvce,&total_core_num);
	if(core_array == NULL) {
		sn_pri(tvce->tdev, SN_ERR, "%s,core_array is NULL\n",
			       __func__);
		goto err;
	}
	for (i = 0; i < total_core_num; i++) {
		id = (core_array + i)->subsys_idx;
		ctype = (core_array + i)->core_type;
		tvce->hantroenc[id].subsys_data.core_info.type_info |=
			(1 << ctype);
		tvce->hantroenc[id].subsys_data.core_info.offset[ctype] =
			(core_array + i)->offset;
		tvce->hantroenc[id].subsys_data.core_info.regSize[ctype] =
			(core_array + i)->reg_size;
		if ((core_array + i)->irq != -1) {
			tvce->hantroenc[id].subsys_data.core_info.irq[ctype] =
				pci_irq_vector(tdev->pdev, (core_array + i)->irq);
			strcpy(tvce->hantroenc[id]
				       .subsys_data.core_info.irq_name[ctype],
			       (core_array + i)->name);
		} else
			tvce->hantroenc[id].subsys_data.core_info.irq[ctype] =
				-1;

		sn_pri(tdev, SN_DBG, "vce:sub:%d, irq:%d \n", i,
		       tvce->hantroenc[id].subsys_data.core_info.irq[ctype]);
	}

	init_waitqueue_head(&tvce->reserve_queue);
	init_waitqueue_head(&tvce->enc_wait_queue);
	init_waitqueue_head(&tvce->reset_core_queue);

	spin_lock_init(&tvce->reserve_lock);
	spin_lock_init(&tvce->irq_lock);

#if S1_VCE
	sys_configure(tdev, 0, SYS_CTL_VCE);
#endif
#if S2_VCE
	sys_configure(tdev, 1, SYS_CTL_VCE);
#endif

	ret = vce_check_id(tdev);
	if (ret < 0)
		goto err;

	/* get the IRQ line */
	for (i = 0; i < tvce->total_subsys_num; i++) {
		if (tvce->hantroenc[i].is_valid == 0)
			continue;
		for (j = 0; j < CORE_MAX; j++) {
			sn_pri(tvce->tdev, SN_DBG, "vce: sub %d, %d,irq:%d \n",
			       i, j,
			       tvce->hantroenc[i].subsys_data.core_info.irq[j]);
			if (tvce->hantroenc[i].subsys_data.core_info.irq[j] > 0) {
				sn_pri(tvce->tdev, SN_DBG,
				       "vce: sub %d, %d,irq:%d\n", i, j,
				       tvce->hantroenc[i]
					       .subsys_data.core_info.irq[j]);
				ret = request_irq(
					tvce->hantroenc[i]
						.subsys_data.core_info.irq[j],
					hantroenc_isr, IRQF_SHARED,
					tvce->hantroenc[i]
						.subsys_data.core_info
						.irq_name[j],
					(void *)&tvce->hantroenc[i]);
				if (ret == -EINVAL) {
					sn_pri(tdev, SN_ERR,
					       "vce: Bad irq number:%d or handler\n",
					       tvce->hantroenc[i]
						       .subsys_data.core_info
						       .irq[j]);
					goto err_irq_free;
				} else if (ret == -EBUSY) {
					sn_pri(tdev, SN_ERR,
					       "vce: irq:%d busy, change your config\n",
					       tvce->hantroenc[i]
						       .subsys_data.core_info
						       .irq[j]);
					goto err_irq_free;
				} else
					sn_pri(tdev, SN_DBG,
					       "vce: irq:%d request irq ok\n",
					       tvce->hantroenc[i]
						       .subsys_data.core_info
						       .irq[j]);
			} else
				sn_pri(tdev, SN_DBG, "vce: IRQ not in use!\n");
		}
	}

	/*disable VCMD gate control:vce & im */
	disable_vcmd_gate(tvce);

	ret = sysfs_create_group(&tdev->misc_dev->this_device->kobj,
				&trans_enc_attribute_group);
	if (ret) {
		sn_pri(tdev, SN_ERR,
			 "encoder: failed to create sysfs device attributes\n");
		goto err_irq_free;
	}
	tvce->loading_timer.expires = jiffies + LOADING_TIME*HZ;
	timer_setup(&tvce->loading_timer, enc_loading_timer_isr, 0);
	add_timer(&tvce->loading_timer);

	sn_pri(tdev, SN_INF, "vce: module initialize done.\n");
	return 0;

err_irq_free:
	for (m = 0; m < i; m++) {
		for (n = 0; n < j; n++) {
			val = tvce->hantroenc[m].subsys_data.core_info.irq[n];
			if (val > 0)
				free_irq(val, (void *)&tvce->hantroenc[m]);
		}
	}
err:
	kfree(tvce);
	sn_pri(tdev, SN_DBG, "vce: module not inserted\n");
	return ret;
}

void vc8000e_release(struct sn_tranx_t *tdev)
{
	u32 i = 0, j;
	struct enc_vce *tvce = tdev->modules[SN_MODULE_VC8000E];

	del_timer_sync(&tvce->loading_timer);
	for (i = 0; i < tvce->total_subsys_num; i++) {
		if (tvce->hantroenc[i].is_valid == 0)
			continue;
		/* free the core IRQ */
		for (j = 0; j < CORE_MAX; j++) {
			if (tvce->hantroenc[i].subsys_data.core_info.irq[j] > 0)
				free_irq(tvce->hantroenc[i]
						 .subsys_data.core_info.irq[j],
					 (void *)&tvce->hantroenc[i]);
		}
	}
#if S1_VCE
	sys_configure_off(tdev, 0, SYS_CTL_VCE);
#endif
#if S2_VCE
	sys_configure_off(tdev, 1, SYS_CTL_VCE);
#endif

	sysfs_remove_group(&tdev->misc_dev->this_device->kobj,
				&trans_enc_attribute_group);
	kfree(tvce);

	sn_pri(tdev, SN_DBG, "vce: module removed\n");
}

long vc8000e_ioctl(struct file *filp, unsigned int cmd, unsigned long arg,
		   struct sn_tranx_t *tdev)
{
	int ret = 0;
	u32 id, io_size, core_info, irq_status;
	struct subsys_core_info in_data;
	struct core_wait_out out;
	struct enc_vce *tvce = tdev->modules[SN_MODULE_VC8000E];
	int vcmd_core_num;

	switch (cmd) {
	case HANTRO_IOCH_GET_VCMD_ENABLE:
#if VCMD_ENABLE_VC8000E
		vcmd_core_num = hantrovcmd_get_codec_core_num(tdev, SN_MODULE_VC8000E);
        vcmd_core_num = (vcmd_core_num ? 1 : 0);
#else
        vcmd_core_num = 0;
#endif
               __put_user(vcmd_core_num, (unsigned long *)arg);
		break;
	case HANTRO_IOCG_HWOFFSET:
		__get_user(id, (u32 *)arg);
		if (id >= tvce->total_subsys_num)
			return -EFAULT;
		__put_user(tvce->hantroenc[id].subsys_data.cfg.base_addr,
			   (unsigned long *)arg);
		break;
	case HANTRO_IOCG_HWIOSIZE:
		__get_user(id, (u32 *)arg);
		if (id >= tvce->total_subsys_num)
			return -EFAULT;
		io_size = tvce->hantroenc[id].subsys_data.cfg.iosize;
		__put_user(io_size, (u32 *)arg);
		break;
	case HANTRO_IOCG_CORE_NUM:
		__put_user(tvce->total_subsys_num, (u32 *)arg);
		break;
	case HANTRO_IOCG_CORE_INFO:
		if (copy_from_user(&in_data, (void *)arg, sizeof(in_data)))
			return -1;
		id = in_data.type_info;
		if (id > tvce->total_subsys_num - 1)
			return -1;
		ret = copy_to_user((void *)arg,
				   &tvce->hantroenc[id].subsys_data.core_info,
				   sizeof(struct subsys_core_info));
		break;
	case HANTRO_IOCH_ENC_RESERVE:
		sn_pri(tdev, SN_DBG, "vce: reserve enc cores\n");
		__get_user(core_info, (u32 *)arg);
		ret = reserve_enc_core(tvce, filp, &core_info);
		if (ret == 0)
			__put_user(core_info, (u32 *)arg);
		break;
	case HANTRO_IOCH_ENC_RELEASE:
		__get_user(core_info, (u32 *)arg);
		sn_pri(tdev, SN_DBG, "vce: release enc core\n");
		release_enc_core(tvce, filp, &core_info);
		break;
	case HANTRO_IOCG_CORE_WAIT:
		__get_user(core_info, (u32 *)arg);
		ret = wait_enc_ready(tvce, filp, &core_info, &irq_status);
		if (ret == 0) {
			__put_user(irq_status, (u32 *)arg);
			return core_info;
		} else {
			__put_user(0, (u32 *)arg);
			return -1;
		}

		break;
	case HANTRO_IOCG_ANYCORE_WAIT:
		memset(&out, 0, sizeof(out));
		ret = wait_enc_any_ready(tvce, filp, &out);
		if (ret == 0)
			if (copy_to_user((void *)arg, &out, sizeof(out)))
				ret = -1;
		break;
	case HANTRO_IOC_SOFT_RESET:
		vc8000e_soft_reset(tdev);
		break;
	case PWRTEST_IOC_VCE_RESET_KEEP:
		vc8000e_subsys_reset_keep(tdev);
		break;
	case PWRTEST_IOC_VCE_RESET_RELEASE:
		vc8000e_subsys_reset_release(tdev);
		break;
	case HANTRO_IOC_SOFT_RESET_SLICE:
		__get_user(core_info, (u32 *)arg);
		sn_pri(tdev, SN_INF, "ready reset core_info %d\n", core_info);
		vc8000e_soft_reset_slice(tdev, core_info);
		break;
	default:
		sn_pri(tdev, SN_ERR, "vce: ioctl cmd:0x%x is error\n", cmd);
	}

	return ret;
}

int vce_dump_regs(struct sn_tranx_t *tdev, struct file * filp)
{
	int ret = 0;
	int i = 0, core_type = 0, offset_count = 0, offset_size = 0, core_name_count = 0;
	unsigned long core_offset = 0;
	u32 reg_value = 0;
	volatile u8 *hw_regs = NULL;
	char write_buf[60] = {0};
	struct enc_vce *tvce = tdev->modules[SN_MODULE_VC8000E];
	if (tdev == NULL || filp == NULL || tvce == NULL)
		return -EFAULT;
	for (i = 0; i < tvce->total_subsys_num; i++) {
		if (!tvce->hantroenc[i].is_valid)
			continue;
		for (core_type = 0; core_type < CORE_MAX; core_type++) {
			if (tvce->hantroenc[i].subsys_data.core_info.regSize[core_type]) {
				core_offset = tvce->hantroenc[i].subsys_data.core_info.offset[core_type];
				offset_size = tvce->hantroenc[i].subsys_data.core_info.regSize[core_type] / 4;
				hw_regs = tvce->hantroenc[i].hwregs;
				if (core_type == CORE_DEC400)
					core_name_count = CORE_VC8000E;
				else
					core_name_count = core_type;
				sn_pri(tvce->tdev, SN_INF, "dump %s %s:\n",
						tvce->hantroenc[i].subsys_data.core_info.irq_name[core_name_count],
						vce_get_coretype_str(core_type));
				sprintf(write_buf, "dump %s %s:\n",
						tvce->hantroenc[i].subsys_data.core_info.irq_name[core_name_count],
						vce_get_coretype_str(core_type));
				ret = write_dump_file(filp, write_buf, strlen(write_buf));
				if (ret < 0)
					return ret;
				reg_value = readl(hw_regs);
				sprintf(write_buf, "%s %s hardware_id=0x%08x]\n",
						tvce->hantroenc[i].subsys_data.core_info.irq_name[core_name_count],
						vce_get_coretype_str(core_type), reg_value);
				ret = write_dump_file(filp, write_buf, strlen(write_buf));
				if (ret < 0)
					return ret;
				reg_value = readl(hw_regs + 320);
				sprintf(write_buf, "%s %s hardware_configure=0x%08x]\n",
						tvce->hantroenc[i].subsys_data.core_info.irq_name[core_name_count],
						vce_get_coretype_str(core_type), reg_value);
				ret = write_dump_file(filp, write_buf, strlen(write_buf));
				if (ret < 0)
					return ret;
				reg_value = readl(hw_regs - 0x1000 + 0x64);
				sprintf(write_buf, "%s %s dec400 status=0x%08x]\n",
						tvce->hantroenc[i].subsys_data.core_info.irq_name[core_name_count],
						vce_get_coretype_str(core_type), reg_value);
				ret = write_dump_file(filp, write_buf, strlen(write_buf));
				if (ret < 0)
					return ret;
				for (offset_count = 0; offset_count < offset_size; offset_count++) {
					reg_value = readl(hw_regs + offset_count * 4);
					sprintf(write_buf, "%s %s reg%d=0x%08x]\n", vce_get_coretype_str(core_type),
							tvce->hantroenc[i].subsys_data.core_info.irq_name[core_name_count], offset_count, reg_value);
					ret = write_dump_file(filp, write_buf, strlen(write_buf));
					if (ret < 0)
						return ret;
				}
			}
			else
				sn_pri(tvce->tdev, SN_DBG, "in vce, the %s is unused\n", vce_get_coretype_str(core_type));
		}
	}
	return 0;
}
