// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * Copyright (C) 2020 Xilinx Inc.
 *
 * This is xilinx scaler driver for Linux.
 * This file provide register operation and initialization,
 * like read/write a register or pull/push a batch of registers.
 */

#include <linux/pci.h>
#include <linux/pagemap.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/kfifo.h>

#include "common.h"
#include "transcoder.h"
#include "hw_monitor.h"
#include "regs.h"

#include "sn_perf.h"

#if defined(TRACE_XAV1_ENC)
#define TRACE trace_printk
#else
#define TRACE(...)
#endif

#define XAV1_ENC_REGMAP_SIZE            (5*1024*1024) //5MB

#define XAV1_HOST_CONTROL_BASE          (0x400000) //(0x12900000 - 0x12500000)
#define XAV1_HOST_CONTROL_AP            (XAV1_HOST_CONTROL_BASE + 0x00)
#define XAV1_HOST_CONTROL_GIE           (XAV1_HOST_CONTROL_BASE + 0x04)
#define XAV1_HOST_CONTROL_ISR           (XAV1_HOST_CONTROL_BASE + 0x0C)

#define XAV1_MBOX_BASE                  (0x401000) //(0x12902000 - 0x12500000)
#define XAV1_MBOX0_DATA                 (XAV1_MBOX_BASE + 0 * 16)
#define XAV1_MBOX0_LAST                 (XAV1_MBOX0_DATA + 4)
#define XAV1_MBOX0_PTR                  (XAV1_MBOX0_LAST + 4)
#define XAV1_MBOX15_DATA                (XAV1_MBOX_BASE + 15 * 16)
#define XAV1_MBOX15_PTR                 (XAV1_MBOX15_DATA + 8)

#define XAV1_SAR_BASE                   (0x402000) //(0x12902000 - 0x12500000)
#define XAV1_SAR_CONTROL0_SET             (XAV1_SAR_BASE + 0x90)
#define XAV1_SAR_CONTROL0_CLR             (XAV1_SAR_BASE + 0x94)
#define XAV1_SAR_CONTROL0_VAL             (XAV1_SAR_BASE + 0x98)
#define XAV1_SAR_CONTROL1_SET             (XAV1_SAR_BASE + 0xD0)
#define XAV1_SAR_CONTROL1_CLR             (XAV1_SAR_BASE + 0xD4)
#define XAV1_SAR_CONTROL1_VAL             (XAV1_SAR_BASE + 0xD8)
#define XAV1_SAR_HOST_INTR_STATUS       (XAV1_SAR_BASE + 0x0A0)
#define XAV1_SAR_FPS_STATUS             (XAV1_SAR_BASE + 0x0A8)
#define XAV1_SAR_FPS_STATUS1            (XAV1_SAR_BASE + 0x0E0)
#define XAV1_SAR_FPS_VERSION            (XAV1_SAR_BASE + 0x0AC)



#define XBIT_MASK(bit)                  ((u32)1<<(bit))

struct xav1_enc_core_desc {
    int slice;
    long base;
    int iosize;
    int irq;
    const char *name;
};

/*here config every core in all subsystems*/
//PF Mode
static struct xav1_enc_core_desc pf_core_array[]= {
    {0, S1_MB_0_AXI_LMB_OFF, XAV1_ENC_REGMAP_SIZE, 18, "xav1_enc_0"},
    {1, S2_MB_0_AXI_LMB_OFF, XAV1_ENC_REGMAP_SIZE, 28, "xav1_enc_1"},
};

//ONE VF Mode
static struct xav1_enc_core_desc onevf_core_array[]= {
    {0, ONE_VF_S1_MB_0_AXI_LMB_OFF, XAV1_ENC_REGMAP_SIZE, 18, "xav1_enc_0"},
    {1, ONE_VF_S2_MB_0_AXI_LMB_OFF, XAV1_ENC_REGMAP_SIZE, 28, "xav1_enc_1"},
};

struct core_cfg {
    int irq;
    unsigned long base_addr;
    volatile u8 *hwreg;
    u32 irq_recvd;
    u32 iosize;
    u32 reg_size;
    u32 *shadow;
    const char *name;
    int irq_count;
};

#define MAX_SLOTS_PER_SLICE 128 // DO *NOT* ADJUST THIS
#define MAX_PENDING_MSGS_PER_SLOT 32
#define MAX_WORDS_PER_MSG 4

struct fifo_msg {
    u32 words[MAX_WORDS_PER_MSG];
};

struct slot {
    wait_queue_head_t wait_queue;
    STRUCT_KFIFO(struct fifo_msg, 128) fifo;
    spinlock_t fifo_lock; // TODO(jkuhn): might not need
    struct file *filp;
    int pendingDelete;
};

struct slice {
    int valid;
    int numChan;
    int nextAvailSlot;
    struct slot slots[MAX_SLOTS_PER_SLICE];
};

struct xav1_enc {
    struct core_cfg subsys[MAX_SLICE_NUM];
    spinlock_t irq_lock;
    spinlock_t reserve_lock;
    wait_queue_head_t wait_queue;
    wait_queue_head_t reserve_queue;
    void *regs_shadow;
    int isCoreIdle[MAX_SLICE_NUM];
    int isCoreLocked[MAX_SLICE_NUM];
    struct sn_tranx_t *tdev;
    struct slice slices[MAX_SLICE_NUM];
    SnPerfHandle perfHandles[SN_PERF_ID_COUNT];
    u32 mb_full[2];
    struct semaphore boot_sem;
    struct file* boot_sem_owner;
    bool fatal_fw_error;
};

static struct xav1_enc_core_desc* get_core_desc(struct xav1_enc *xav1e, int *num_cores)
{
    struct xav1_enc_core_desc *cp = NULL;
    int ncores = 0;

    switch(xav1e->tdev->pf_vf_mode)
    {
        case PF_MODE:
            if (xav1e->tdev->vf_index != PF_INDEX) {
                sn_pri(xav1e->tdev, SN_ERR, "%s,vf_index incorrect in PF_MODE\n", __func__);
            } else {
                ncores = sizeof(pf_core_array)/sizeof(pf_core_array[0]);
                cp     = pf_core_array;
            }
            break;

        case ONE_VF_MODE:
            if (xav1e->tdev->vf_index != VF1_INDEX) {
                sn_pri(xav1e->tdev, SN_ERR, "%s,vf_index incorrect in ONE_VF_MODE\n", __func__);
            } else {
                ncores = sizeof(onevf_core_array)/sizeof(onevf_core_array[0]);
                cp     = onevf_core_array;
            }
            break;

        default:
            sn_pri(xav1e->tdev, SN_ERR, "%s,incorrect pf_vf_mode (%d)\n", __func__, xav1e->tdev->pf_vf_mode);
            break;
    }

    *num_cores = ncores;
    return cp;
}


static int init_core_array(struct xav1_enc *xav1e)
{
    int i, irq;
    u32 total_io_size = 0;
    struct xav1_enc_core_desc* cd = NULL;
    int num_cores = 0;

    //get core description table for pf/vf
    cd = get_core_desc(xav1e, &num_cores);
    if (!num_cores) {
        sn_pri(xav1e->tdev, SN_ERR, "failed to get num_cores\n");
        return 0;
    }

    for (i = 0; i < num_cores; i++) {
        xav1e->subsys[i].base_addr = (cd+i)->base;
        xav1e->subsys[i].iosize    = (cd+i)->iosize;
        irq                        = (cd+i)->irq;
        if (irq != -1)
            xav1e->subsys[i].irq   = pci_irq_vector(xav1e->tdev->pdev, irq);

        total_io_size              += xav1e->subsys[i].iosize;
        xav1e->subsys[i].hwreg     = 0;
        xav1e->subsys[i].name      = (cd+i)->name;

        sn_pri(xav1e->tdev, SN_DBG, "xav1_enc:slice[%d]: base=%x  iosize=%d  tot_size=%d  core_irq id: %d  pci_irq: %d\n",
                i, xav1e->subsys[i].base_addr, xav1e->subsys[i].iosize, total_io_size,
                irq, xav1e->subsys[i].irq);
    }

    return total_io_size;
}

static int reserve_io(struct xav1_enc *xav1e)
{
    int i;

    for (i = 0; i < MAX_SLICE_NUM; i++) {
        if (xav1e->subsys[i].iosize) {
            sn_pri(xav1e->tdev, SN_DBG, "xav1_enc:slice[%d]: bar2_virt=0x%x  ipbase=0x%lx, iosize=%d\n", i,
                    xav1e->tdev->bar2_virt, xav1e->subsys[i].base_addr, xav1e->subsys[i].iosize);

            xav1e->subsys[i].hwreg = xav1e->tdev->bar2_virt + xav1e->subsys[i].base_addr;

            if (xav1e->subsys[i].hwreg == NULL) {
                sn_pri(xav1e->tdev, SN_ERR, "xav1_enc: failed to ioremap HW %d regs\n", i);
                return -EBUSY;
            }
            sn_pri(xav1e->tdev, SN_DBG, "xav1_enc:slice[%d]: hwreg(bar2_virt+base)=0x%x  reg[0]=0x%08x\n",
                i, xav1e->subsys[i].hwreg, readl(xav1e->subsys[i].hwreg));
        }
    }
    return 0;
}

static int check_slice(struct xav1_enc *xav1e, unsigned slice) {
    if (slice >= MAX_SLICE_NUM || !xav1e->subsys[slice].base_addr || !xav1e->slices[slice].valid) {
        TRACE("Invalid slice: %u\n", slice);
        return -1;
    }
    return 0;
}

static long xav1_enc_write_regs(struct xav1_enc *xav1e, struct core_desc *core)
{
    long ret;
    u32 i, reg_cnt;
    u32 id = core->id;
    struct reg_desc *regs_info;

    if (check_slice(xav1e, id)) {
        sn_pri(xav1e->tdev, SN_ERR, "xav1_enc: %s failed core_id: %d base_addr: %uld\n",
               __func__, id,xav1e->subsys[id].base_addr);
        return -EINVAL;
    }

    regs_info = (struct reg_desc *)xav1e->subsys[id].shadow;

    ret = copy_from_user(regs_info, core->regs, core->size);
    if (ret) {
        sn_pri(xav1e->tdev, SN_ERR, "xav1_enc: %s failed %ld\n", __func__, ret);
        return -EFAULT;
    }

    reg_cnt = core->size / sizeof(struct reg_desc);
    for (i = 0; i < reg_cnt; i++) {
        writel(regs_info[i].val,  xav1e->subsys[id].hwreg + regs_info[i].id);
    }
    return 0;
}

static long xav1_enc_read_regs(struct xav1_enc *xav1e, struct core_desc *core)
{
    long ret;
    u32 id = core->id;
    u32 i, reg_cnt;
    struct reg_desc *regs_info;

    if (check_slice(xav1e, id)) {
        sn_pri(xav1e->tdev, SN_ERR, "xav1_enc: %s failed core_id: %d base_addr: %uld\n",
               __func__, id,xav1e->subsys[id].base_addr);
        return -EINVAL;
    }

    regs_info = (struct reg_desc *)xav1e->subsys[id].shadow;

    ret = copy_from_user(regs_info, core->regs, core->size);
    if (ret) {
        sn_pri(xav1e->tdev, SN_ERR, "xav1_enc: %s failed %ld\n", __func__, ret);
        return -EFAULT;
    }

    reg_cnt = core->size / sizeof(struct reg_desc);
    /* read specific registers from hardware */
    for (i = 0; i < reg_cnt; i++) {
        regs_info[i].val = readl(xav1e->subsys[id].hwreg + regs_info[i].id);
    }

    /* put registers to user space*/
    ret = copy_to_user(core->regs, regs_info, core->size);
    if (ret) {
        sn_pri(xav1e->tdev, SN_ERR, "xav1_enc: %s failed %ld\n", __func__, ret);
        return -EFAULT;
    }

    return 0;
}

static int xav1_enc_alloc_slot(struct xav1_enc *Enc, struct file *filp, void __user *user) {
    struct slice *slice;
    struct slot *slot;
    struct slot *end;
    struct xav1_slot slotInfo;
    unsigned long flags;

    if (copy_from_user(&slotInfo, user, sizeof(slotInfo))) {
        return -EFAULT;
    }
    if (check_slice(Enc, slotInfo.slice)) {
        return -EINVAL;
    }
    slice = Enc->slices + slotInfo.slice;
    spin_lock_irqsave(&Enc->reserve_lock, flags);
    if (!slice->numChan) {
        spin_unlock_irqrestore(&Enc->reserve_lock, flags);
        return -ENOMEM;
    }
    slot = slice->slots + slice->nextAvailSlot;
    end = slot;
    slotInfo.slot = slice->nextAvailSlot;
    do {
        if (!slot->filp) {
            slot->filp = filp;
            slot->pendingDelete = 0;
            slice->nextAvailSlot = (slotInfo.slot + 1) % MAX_SLOTS_PER_SLICE;
            --slice->numChan;
            spin_unlock_irqrestore(&Enc->reserve_lock, flags);
            init_waitqueue_head(&slot->wait_queue);
            spin_lock_init(&slot->fifo_lock);
            INIT_KFIFO(slot->fifo);
            TRACE("Allocated slot: %u:%u [%u] for %p\n", slotInfo.slice, slotInfo.slot, slice->numChan, filp);
            return copy_to_user(user, &slotInfo, sizeof(slotInfo)); // would leak if this fails
        }
        if (++slotInfo.slot == MAX_SLOTS_PER_SLICE) {
            slotInfo.slot = 0;
            slot = slice->slots;
        } else {
            ++slot;
        }
    } while (slot != end);
    // should not be possible
    spin_unlock_irqrestore(&Enc->reserve_lock, flags);
    printk(KERN_CRIT "No free slots (should be)\n");
    return -EINVAL;
}

static int xav1_enc_send_msg_raw(struct xav1_enc *Enc, struct xav1_msg *Msg, int Lock)
{
    volatile u8* hwregs;
    int i;
    unsigned long flags;
    hwregs = Enc->subsys[Msg->slice].hwreg;
    if (Lock) {
        spin_lock_irqsave(&Enc->reserve_lock, flags);
    }
    {
        int rawVal = readl(hwregs + XAV1_MBOX0_PTR);
        int writeIdx = (rawVal >> 8) & 127;
        int readIdx  = (rawVal >> 20) & 127;
        int avail = ((readIdx + 128 - writeIdx) & 127) + 128 * !!(rawVal & 1);
        TRACE("message: %08x %d %d %d %d\n", rawVal, writeIdx, readIdx, avail, Msg->count);
        if (avail < Msg->count) {
            if (Lock) {
                spin_unlock_irqrestore(&Enc->reserve_lock, flags);
            }
            TRACE("No room for message: %08x %d %d %d %d\n", rawVal, writeIdx, readIdx, avail, Msg->count);
            ++Enc->mb_full[Msg->slice];
            return -EAGAIN;
        }
    }
    for (i = 0; i < Msg->count - 1; ++i) {
        writel(Msg->content[i], hwregs + XAV1_MBOX0_DATA);
    }
    writel(Msg->content[i], hwregs + XAV1_MBOX0_LAST);
    if (Lock) {
        spin_unlock_irqrestore(&Enc->reserve_lock, flags);
    }
    return 0;
}

static int xav1_enc_free_slot(struct xav1_enc *Enc, struct file *filp, void __user *user) {
    struct xav1_slot slotInfo;
    struct xav1_msg msg;
    struct slice *slice;
    struct slot *slot;
    unsigned long flags;
    if (copy_from_user(&slotInfo, user, sizeof(slotInfo))) {
        return -EFAULT;
    }
    if (slotInfo.slot >= MAX_SLOTS_PER_SLICE || check_slice(Enc, slotInfo.slice)) {
        return -EINVAL;
    }
    slice = Enc->slices + slotInfo.slice;
    slot = slice->slots + slotInfo.slot;
    msg.slice = slotInfo.slice;
    msg.count = 1;
    msg.content[0] = 255 | (slotInfo.slot << 10);
    spin_lock_irqsave(&Enc->reserve_lock, flags);
    if (slot->filp != filp || slot->pendingDelete) { 
        // A reset may have revoked ownership of this slot, or maybe we're dealing with an ill-behaved application.
        spin_unlock_irqrestore(&Enc->reserve_lock, flags);
        TRACE("Skipping %p %u\n", slot->filp, slot->pendingDelete);
        return -EINVAL;
    }
    // TODO(jkuhn): slot destruction (pending?)
    if (xav1_enc_send_msg_raw(Enc, &msg, 0) < 0) { // no lock: We're already holding reserve_lock to exclude the possibility of reset.
        spin_unlock_irqrestore(&Enc->reserve_lock, flags);
        return -EAGAIN;
    }
    slot->pendingDelete = 1;
    spin_unlock_irqrestore(&Enc->reserve_lock, flags);
    return 0;
}

static int xav1_enc_send_msg(struct xav1_enc *Enc, void __user *user)
{
    struct xav1_msg msg;
    if (copy_from_user(&msg, user, sizeof(msg))) {
        return -EFAULT;
    }
    if (check_slice(Enc, msg.slice) || msg.count > 4) {
        TRACE("Invalid Message: slice: %d count: %d\n", msg.slice, msg.count);
        return -EINVAL;
    }
    return xav1_enc_send_msg_raw(Enc, &msg, 1);
}

static int  xav1_enc_wait_msg(struct xav1_enc *Enc, struct file *filp, void __user *user)
{
    struct xav1_wait_msg msg;
    struct slice* slice;
    struct slot* slot;
    int ret;
    if (copy_from_user(&msg, user, sizeof(msg))) {
        return -EFAULT;
    }
    if (msg.slot >= MAX_SLOTS_PER_SLICE || check_slice(Enc, msg.slice)) {
        return -EINVAL;
    }
    slice = Enc->slices + msg.slice;
    slot = slice->slots + msg.slot;
    if (slot->filp != filp) {
        return -EINVAL;
    }
    ret = wait_event_interruptible_timeout(slot->wait_queue, !kfifo_is_empty(&slot->fifo), msg.timeout_ms * HZ / 1000);
    if (ret < 0) {
        return ret;
    }
    msg.timeout_ms = ret * 1000 / HZ;
    if (ret) {
        int i;
        struct fifo_msg val;
        if (!kfifo_get(&slot->fifo, &val)) {
            TRACE("FIFO was empty!\n");
            return -EINVAL;
        }
        for (i = 0; i < MAX_WORDS_PER_MSG; ++i) {
            msg.content[i] = val.words[i];
        }
    }
    return copy_to_user(user, &msg, sizeof(msg));
}

static int process_host_messages(struct xav1_enc *Enc, int Slice)
{
    // TODO(jkuhn) error handling and recovery
    volatile u8 *hwregs = Enc->subsys[Slice].hwreg;
    struct fifo_msg msg;
    u32 header;
    int numWords;
    int slotNum;
    int arg;
    msg.words[0] = readl(hwregs + XAV1_MBOX15_DATA);
    header = msg.words[0];
    numWords = (header >> 8) & 3;
    if (numWords) {
        int i;
        for (i = 1; i < numWords + 1; ++i) {
            msg.words[i] = readl(hwregs + XAV1_MBOX15_DATA);
        }
    }
    if ((header & 255) == 252) {
        int cpu = header >> 30;
        if (numWords) {
            int index = (header >> 10) & 0xfffff;
            int i;
            for (i = 1; i < numWords + 1; ++i) {
                printk(KERN_CRIT "[BT] %08x %d Device %s Slice %d Cpu %d\n", msg.words[i], index + i - 1, Enc->tdev->dev_name, Slice, cpu);
            }
        } else {
            int error = (header >> 10) & 0xfffff;
            Enc->fatal_fw_error = true;
            printk(KERN_CRIT "FATAL FPS ERROR: %08x Device %s Slice %d Cpu %d\n", error, Enc->tdev->dev_name, Slice, cpu);
        }
        return 0;
    }
    if ((header & 255) == 253) {
        u32 ipId = (header >> 11) & 127;
        u32 load = (header >> 18);
        if (ipId >= SN_PERF_ID_COUNT) {
          TRACE("INGORING UNKNOWN IP: %u", ipId);
          return 0;
        }
        sn_perf_load(Enc->perfHandles[ipId], load);
        return 0;
    }
    if ((header & 255) == 254) {
        sn_perf_event event;
        TRACE("%08x\n", header);
        event.bits = header >> 10;
        event.hwTs = msg.words[1];
        event.payload = msg.words[2];
        event.name = msg.words[3];
        TRACE("%u.%03u.%02u %u %9u %08x %.4s\n", event.slice, event.ipId, event.unitId, event.type, event.hwTsValid, event.hwTs, event.namec);
        if (event.ipId >= SN_PERF_ID_COUNT) {
          TRACE("INGORING UNKNOWN IP: %u", event.ipId);
          return 0;
        }
        sn_perf_record(Enc->perfHandles[event.ipId], &event);
        return 0;
    }
    slotNum = (header >> 10) & 0xff;
    arg = (header >> 18);
    TRACE("Got msg %d for %d + %d [%x]\n", header & 255, slotNum, numWords, header);
    if (slotNum < MAX_SLOTS_PER_SLICE) {
        struct slot* slot = Enc->slices[Slice].slots + slotNum;
        if (!slot->filp) {
            TRACE("DROPPING message for unallocated slot %u:%u\n", Slice, slotNum);
            return 0;
        }
        if (slot->pendingDelete) {
            if ((header & 255) == 255) { // teardown
                void* filp;
                struct slice* slice = Enc->slices + Slice;
                unsigned long flags;
                spin_lock_irqsave(&Enc->reserve_lock, flags);
                filp = slot->filp;
                slot->filp = NULL;
                ++slice->numChan;
                spin_unlock_irqrestore(&Enc->reserve_lock, flags);
                TRACE("Freed slot: %u:%u [%u] for %p\n", Slice, slotNum, slice->numChan, filp);
                return 0;
            }
            TRACE("Ignoring message for slot %u:%u pending delete\n", Slice, slotNum);
            return 0;
        }
        if (!kfifo_put(&slot->fifo, msg)) {
            // TODO(jkuhn): How to communicate this error state to application?
            TRACE("DROPPING message for full slot %u:%u\n", Slice, slotNum);
        } else {
            wake_up_all(&slot->wait_queue);
        }
    }
    return 0;
}

static irqreturn_t xav1_enc_isr(int irq, void *dev_id)
{
    unsigned long flags;
    volatile u8 *hwregs;
    u32 irq_status;
    int slice = 0;
    struct xav1_enc *xav1e = (struct xav1_enc *) dev_id;
    u32  int_shadow_status_off = S1_INT_SD_XENC_STUS;
    u32 int_shadow_isr_off = S1_INT_SD_XENC_ISR;

    TRACE("ISR\n");
    spin_lock_irqsave(&xav1e->irq_lock, flags);
    //determine slice index from irq#
    if(irq == xav1e->subsys[0].irq) {
        slice = 0;
        int_shadow_status_off = S1_INT_SD_XENC_STUS;
        int_shadow_isr_off = S1_INT_SD_XENC_ISR;
    } else if(irq == xav1e->subsys[1].irq) {
        slice = 1;
        int_shadow_status_off = S2_INT_SD_XENC_STUS;
        int_shadow_isr_off = S2_INT_SD_XENC_ISR;
    } else {
        sn_pri(xav1e->tdev, SN_ERR, "xav1: error:: irq %d not supported \n", irq);
    }

    hwregs = xav1e->subsys[slice].hwreg;

    /* Semaphore HOST interrupt status register read */
    irq_status = readl(hwregs + XAV1_SAR_HOST_INTR_STATUS);
    //check if Host IRQ
    TRACE("irq_status (shadow): %08x read: %08x\n", irq_status, readl(hwregs + XAV1_SAR_HOST_INTR_STATUS));
    if (irq_status & XBIT_MASK(29)) {
        process_host_messages(xav1e, slice);
    }

    spin_unlock_irqrestore(&xav1e->irq_lock, flags);

    return IRQ_HANDLED;
}

#if SUB_SYS_VCE != 1 /* assume legoVSI FPGA which doesn't have xav1 HW */
static void internal_configure_on(struct xav1_enc *xav1e)
{
    int i;
    u32 val;
    volatile u8 *base_addr=NULL;
    unsigned int loop;

    for (i = 0; i < MAX_SLICE_NUM; i++) {
        base_addr =  xav1e->tdev->bar2_virt + core_array[i].base;
		loop = 1000;
        val = readl(base_addr + XAV1_SAR_CONTROL0_CLR);
        // writel(val|0x1000,base_addr + THS0_POWER_CTRL_SET);
        // printk("%s %d, set reg base=0x%x, poll reg base=0x%x\n",__func__,__LINE__,(base_addr+THS0_POWER_CTRL_SET),(base_addr+THS0_POWER_CTRL_POLL));
        writel(0x1000,base_addr + XAV1_SAR_CONTROL0_CLR);

        val = readl(base_addr + XAV1_SAR_FPS_STATUS);
        while ((val&0x3000000) != 0x0) {
            val = readl(base_addr + XAV1_SAR_FPS_STATUS);
            msleep(1);
            loop--;
            if (!loop) {
                sn_pri(xav1e->tdev,SN_ERR,"xav1 internal power on configuration error!\n");
                break;
            }
        }
    }
}

static void internal_configure_off(struct xav1_enc *xav1e)
{
    int i;
    u32 val;
    volatile u8 *base_addr=NULL;
    unsigned int loop;

    for (i = 0; i < MAX_SLICE_NUM; i++) {
		loop = 1000;
        base_addr =  xav1e->tdev->bar2_virt + core_array[i].base;
        val = readl(base_addr + XAV1_SAR_CONTROL0_SET);
        writel(0x1000,base_addr + XAV1_SAR_CONTROL0_SET);

        val = readl(base_addr + XAV1_SAR_FPS_STATUS);
        while ((val&0x3000000) != 0x3000000) {
            val = readl(base_addr + XAV1_SAR_FPS_STATUS);
            msleep(1);
            loop--;
            if (!loop) {
                sn_pri(xav1e->tdev,SN_ERR,"xav1 internal power off  configuration error!\n");
                break;
            }
        }
    }
}
#endif

#if 0
static long xav1_release_core(struct xav1_enc *xav1e, u32 slice)
{
    if (slice >= MAX_SLICE_NUM)
    {
        sn_pri(xav1e->tdev, SN_ERR, "xav1: %s invalid slice id %d\n", __func__, slice);
        return -EINVAL;
    }

    // update core status to idle and clear lock state
    spin_lock(&xav1e->reserve_lock);
    xav1e->isCoreIdle[slice] = 1;
    xav1e->isCoreLocked[slice] = 0;
    spin_unlock(&xav1e->reserve_lock);

    // signal waiting processes
    wake_up_interruptible_all(&xav1e->reserve_queue);

    return 0;
}
#endif

// reclaim all allocated slots and prevent new allocations
static int cancel_slot_allocation(struct xav1_enc* Enc, int SliceNo)
{
    unsigned long flags;
    if (SliceNo < 0 || SliceNo >= MAX_SLICE_NUM) {
        return -EINVAL;
    }
    spin_lock_irqsave(&Enc->reserve_lock, flags);
    {
        struct slice* slice = Enc->slices + SliceNo;
        struct slot* slot = slice->slots;
        int j;
        slice->numChan = 0;
        for (j = 0; j < MAX_SLOTS_PER_SLICE; ++j, ++slot) {
            slot->filp = NULL;
        }
    }
    spin_unlock_irqrestore(&Enc->reserve_lock, flags);
    return 0;
}

// allow allocations again after cancel_slot_allocation
static int begin_slot_allocation(struct xav1_enc* Enc, int SliceNo)
{
    unsigned long flags;
    if (SliceNo < 0 || SliceNo >= MAX_SLICE_NUM) {
        return -EINVAL;
    }
    spin_lock_irqsave(&Enc->reserve_lock, flags);
    Enc->slices[SliceNo].numChan = MAX_SLOTS_PER_SLICE;
    spin_unlock_irqrestore(&Enc->reserve_lock, flags);
    return 0;
 }

int xav1_whole_subsys_soft_reset(struct xav1_enc *xav1e, int slice)
{
    volatile u8 *hwregs = NULL;
    if (slice < 0 || slice >= MAX_SLICE_NUM)
    {
        return -1;
    }
    hwregs = xav1e->subsys[slice].hwreg;

    sys_xav1_whole_subsys_soft_reset(xav1e->tdev, slice);
    cancel_slot_allocation(xav1e, slice);
    sys_xav1_whole_subsys_release_soft_reset(xav1e->tdev, slice);
    writel(1 << 29, hwregs + XAV1_SAR_CONTROL0_SET); // enable MBOX15 host interrupt
    begin_slot_allocation(xav1e, slice);
    return 0;
}

static int enable_mbox15_host_intr(struct xav1_enc *xav1e, int slice)
{
    volatile u8 *hwregs = NULL;

    if (slice < 0 || slice >= MAX_SLICE_NUM)
    {
        return -1;
    }
    hwregs = xav1e->subsys[slice].hwreg;

    writel(1 << 29, hwregs + XAV1_SAR_CONTROL0_SET);

    return 0;
}

extern unsigned long s1_start, s2_start; // TODO(jkuhn)

static int init_slices(struct xav1_enc* Enc)
{
    struct slice* slice = Enc->slices;
    int i;
    TRACE("init_slices\n");
    for (i = 0; i < MAX_SLICE_NUM; ++i, ++slice) {
        struct slot* slot = slice->slots;
        volatile u8 *hwregs = Enc->subsys[i].hwreg;
        int j;
        slice->numChan = MAX_SLOTS_PER_SLICE;
        slice->nextAvailSlot = 1;
        for (j = 0; j < MAX_SLOTS_PER_SLICE; ++j, ++slot) {
            slot->filp = NULL;
        }
        slice->valid = readl(hwregs + XAV1_MBOX0_PTR) != 0; // TODO(jkuhn): robust?
        writel(1 << 29, hwregs + XAV1_SAR_CONTROL0_SET); // enable MBOX15 host interrupt
        writel(0, hwregs + XAV1_MBOX15_PTR); // reset mailboxes
        writel(0, hwregs + XAV1_MBOX0_PTR); // reset mailboxes
    }
    return 0;
}

static int xav1_enc_perf(struct sn_tranx_t* tdev, __u32 ipId, __u32 cmd, __u32 arg) {
    struct xav1_enc* xav1e = tdev->modules[SN_MODULE_XAV1_ENC];
    struct xav1_msg msg;
    struct slice* slice = xav1e->slices;
    int i;
    for (i = 0; i < MAX_SLICE_NUM; ++i, ++slice) {
        if (!slice->valid) {
            continue;
        }
        msg.slice = i;
        msg.count = 1;
        msg.content[0] = 254 | (ipId << 10) | (cmd << 17) | (arg << 19);
        xav1_enc_send_msg_raw(xav1e, &msg, 1); // TODO(jkuhn): spin???
        TRACE("%u %u %u %u\n", i, ipId, cmd, arg);
    }
    return 0;
}

static ssize_t enable_load_store(struct device* Dev, struct device_attribute* Attr, const char* Buf, size_t Count)
{
    struct sn_misc_tdev* mtdev = dev_get_drvdata(Dev);
    struct sn_tranx_t* tdev = mtdev->tdev;
    struct xav1_enc* xav1e = tdev->modules[SN_MODULE_XAV1_ENC];
    struct xav1_msg msg;
    if (Count != 2 || (Buf[0] != '0' && Buf[0] != '1')) {
        sn_pri(tdev, SN_ERR, "Must be 0 or 1 only\n");
        return -EINVAL;
    }
    msg.slice = 0; // send to both slices;
    msg.count = 1;
    msg.content[0] = 251 | ((Buf[0] - '0') << 10);
    xav1_enc_send_msg_raw(xav1e, &msg, 1);
    msg.slice = 1;
    xav1_enc_send_msg_raw(xav1e, &msg, 1);
    return Count;
}

static ssize_t mb_full_show(struct device* Dev, struct device_attribute* Attr, char* Buf)
{
    struct sn_misc_tdev* mtdev = dev_get_drvdata(Dev);
    struct sn_tranx_t* tdev = mtdev->tdev;
    struct xav1_enc* xav1e = tdev->modules[SN_MODULE_XAV1_ENC];
    return sprintf(Buf, "%u %u\n", xav1e->mb_full[0], xav1e->mb_full[1]);
}

static DEVICE_ATTR_WO(enable_load);
static DEVICE_ATTR_RO(mb_full);

static struct attribute* trans_xav1_sysfs_entries[] = {
    &dev_attr_enable_load.attr, &dev_attr_mb_full.attr, NULL
};

static struct attribute_group trans_xav1_attribute_group = {
    .name = NULL,
    .attrs = trans_xav1_sysfs_entries,
};

int xav1_get_register(struct sn_tranx_t *tdev, u32 slice_num, u32 reg)
{
    struct xav1_enc* xav1e = tdev->modules[SN_MODULE_XAV1_ENC];
    return readl(xav1e->subsys[slice_num].hwreg + reg);
}

int xav1_soft_reset(struct sn_tranx_t *tdev)
{
    struct xav1_enc* xav1e = tdev->modules[SN_MODULE_XAV1_ENC];

    if (xav1_whole_subsys_soft_reset(xav1e, 0) != 0) {
        return -1;
    }

    if (xav1_whole_subsys_soft_reset(xav1e, 1) !=0) {
        return -1;
    }

    return 0;
}

int xav1_enc_init(struct sn_tranx_t *tdev)
{
    int i, result;
    struct xav1_enc *xav1e;
    int total_iosize;
    void *tmp_mem;

    xav1e = (struct xav1_enc *)kzalloc(sizeof(*xav1e), GFP_KERNEL);
    if (!xav1e) {
        sn_pri(tdev, SN_ERR, "xav1_enc: kmalloc failed\n");
        return -ENOMEM;
    }
    tdev->modules[SN_MODULE_XAV1_ENC] = xav1e;
    xav1e->tdev = tdev;

    total_iosize = init_core_array(xav1e);
    if (!total_iosize) {
        sn_pri(xav1e->tdev, SN_ERR, "xav1_enc: total core count is zero\n");
        goto err_free_xav1e;
    }

    for (i = 0; i < MAX_SLICE_NUM; i++) {
        //sys_configure(tdev, i, SYS_CTL_XENC);
        //sys_configure(tdev, i, SYS_CTL_XFPS);
        //sys_configure(tdev, i, SYS_CTL_XAV1);
        sys_configure_ext(tdev, i);
    }

#if SUB_SYS_VCE != 1 /* assume legoVSI FPGA which doesn't have xav1 HW */
    internal_configure_on(xav1e);
#endif

    spin_lock_init(&xav1e->irq_lock);
    spin_lock_init(&xav1e->reserve_lock);

    result = sysfs_create_group(&tdev->misc_dev->this_device->kobj, &trans_xav1_attribute_group);
    if (result) {
        sn_pri(tdev, SN_ERR, "xav1: failed to create sysfs device attributes\n");
        goto err_free_xav1e;
    }

    result = reserve_io(xav1e);
    if(result < 0) {
        sn_pri(xav1e->tdev, SN_ERR, "xav1_enc: reserve io failed\n");
        goto err_free_xav1e;
    }

    tmp_mem = vzalloc(total_iosize);
    if (!tmp_mem) {
        sn_pri(xav1e->tdev, SN_ERR, "xav1_enc: malloc regs mem failed.\n");
        goto err_free_xav1e;
    }

    xav1e->regs_shadow = tmp_mem;
    for (i = 0; i < MAX_SLICE_NUM; i++) {
        if (xav1e->subsys[i].iosize) {
            xav1e->subsys[i].shadow = tmp_mem;
            tmp_mem += xav1e->subsys[i].iosize;
        }
    }

    for (i = 0; i < MAX_SLICE_NUM; i++) {
        if(xav1e->subsys[i].irq > 0) {
            result = request_irq(xav1e->subsys[i].irq, xav1_enc_isr,
                                 IRQF_SHARED, xav1e->subsys[i].name, (void *)xav1e);
            if(result != 0) {
                if(result == -EINVAL) {
                    sn_pri(tdev, SN_ERR,
                        "xav1_enc: Bad IRQ:%d or handler\n",
                        xav1e->subsys[i].irq);
                } else if(result == -EBUSY) {
                    sn_pri(tdev, SN_ERR,
                        "xav1_enc: IRQ:%d busy\n",
                        xav1e->subsys[i].irq);
                }
                goto err_free_irq;
            } else {
                sn_pri(tdev, SN_INF, "xav1_enc:slice[%d]: pci_irq %d registered!\n", i, xav1e->subsys[i].irq);
            }
        } else
            sn_pri(tdev, SN_INF, "xav1_enc: IRQ:%d not in use!\n", i);
    }
    init_slices(xav1e);
    xav1e->perfHandles[SN_PERF_ID_VPP] = sn_perf_register(tdev, "enc_vpp", SN_PERF_ID_VPP, xav1_enc_perf);
    xav1e->perfHandles[SN_PERF_ID_ME] = sn_perf_register(tdev, "enc_me", SN_PERF_ID_ME, xav1_enc_perf);
    xav1e->perfHandles[SN_PERF_ID_VCE] = sn_perf_register(tdev, "enc_vce", SN_PERF_ID_VCE, xav1_enc_perf);
    xav1e->perfHandles[SN_PERF_ID_AV1] = sn_perf_register(tdev, "enc_av1", SN_PERF_ID_AV1, xav1_enc_perf);
    xav1e->perfHandles[SN_PERF_ID_SYSTEMCPU] = sn_perf_register(tdev, "enc_systemcpu", SN_PERF_ID_SYSTEMCPU, xav1_enc_perf);
    xav1e->perfHandles[SN_PERF_ID_INLINECPU] = sn_perf_register(tdev, "enc_inlinecpu", SN_PERF_ID_INLINECPU, xav1_enc_perf);
    xav1e->perfHandles[SN_PERF_ID_LEGOCPU0] = sn_perf_register(tdev, "enc_legocpu0", SN_PERF_ID_LEGOCPU0, xav1_enc_perf);
    xav1e->perfHandles[SN_PERF_ID_LEGOCPU1] = sn_perf_register(tdev, "enc_legocpu1", SN_PERF_ID_LEGOCPU1, xav1_enc_perf);
    xav1e->perfHandles[SN_PERF_ID_CUTREE] = sn_perf_register(tdev, "enc_cutree", SN_PERF_ID_CUTREE, xav1_enc_perf);

    sema_init(&xav1e->boot_sem, 1);

    /* load encoder firmware to FPS sub system */
    result = load_fps_subsystem(tdev);
    if(result) {
        goto err_free_irq;
    }

    xav1e->fatal_fw_error = false;

    sn_pri(tdev, SN_INF, "xav1_enc: module initialization done\n");
    return 0;

err_free_irq:
    for (i= 0; i < MAX_SLICE_NUM; i++) {
        if(xav1e->subsys[i].irq > 0)
          free_irq(xav1e->subsys[i].irq, (void *)xav1e);
    }
    vfree(xav1e->regs_shadow);

err_free_xav1e:
    kfree(xav1e);
    sn_pri(tdev, SN_ERR, "xav1_enc: module not inserted\n");
    return result;
}

void xav1_enc_release(struct sn_tranx_t *tdev)
{
    struct xav1_enc *xav1e = tdev->modules[SN_MODULE_XAV1_ENC];
    int i;

    /* free the IRQ */
    for (i = 0; i < MAX_SLICE_NUM; i++) {
        if(xav1e->subsys[i].irq > 0) {
            xav1_whole_subsys_soft_reset(xav1e, i);
            free_irq(xav1e->subsys[i].irq, (void *)xav1e);
        }
    }
#if SUB_SYS_VCE != 1
    internal_configure_off(xav1e);
#endif

    for (i = 0; i < MAX_SLICE_NUM; i++) {
        sys_configure_off(tdev, i, SYS_CTL_XENC);
        sys_configure_off(tdev, i, SYS_CTL_XFPS);
        sys_configure_off(tdev, i, SYS_CTL_XAV1);
    }
    vfree(xav1e->regs_shadow);

    sysfs_remove_group(&tdev->misc_dev->this_device->kobj, &trans_xav1_attribute_group);

    sn_pri(tdev, SN_DBG, "xav1_enc: remove module done.\n");
}

long xav1_enc_ioctl(struct file *filp,
                    unsigned int cmd,
                    unsigned long argp,
                    struct sn_tranx_t *tdev)
{
    long ret = 0;
    u32 slice;
    struct core_desc core_des;
    struct xav1_enc *xav1e = tdev->modules[SN_MODULE_XAV1_ENC];
    void __user *arg = (void __user *)argp;

    if (xav1e->fatal_fw_error) {
        return -ENOTRECOVERABLE;
    }

    switch(cmd)
    {
    case XAV1_ENC_IOCTL_READ_REG:
    case XAV1_ENC_IOCTL_WRITE_REG:
        ret = copy_from_user(&core_des, (void *)arg, sizeof(struct core_desc));
        if (ret) {
            sn_pri(tdev, SN_ERR, "xbar: ioctl copy_from_user failed %d\n", ret);
            return -EFAULT;
        }
        break;
    case PWRTEST_XAV1_RESET_KEEP:
    case PWRTEST_XAV1_RESET_RELEASE:
    case XAV1_ENC_IOCTL_TWO_SLICE_SOFT_RESET:
    default:
           break;
    }

    switch (cmd) {
    case XAV1_ENC_IOCTL_READ_REG:
        ret = xav1_enc_read_regs(xav1e, &core_des);
        break;
    case XAV1_ENC_IOCTL_WRITE_REG:
        ret = xav1_enc_write_regs(xav1e, &core_des);
        break;
    case XAV1_ENC_IOCTL_ALLOC_SLOT:
        ret = xav1_enc_alloc_slot(xav1e, filp, arg);
        break;
    case XAV1_ENC_IOCTL_FREE_SLOT:
        ret = xav1_enc_free_slot(xav1e, filp, arg);
        break;
    case XAV1_ENC_IOCTL_SEND_MSG: // TODO(jkuhn) - spin or EAGAIN??
        ret = xav1_enc_send_msg(xav1e, arg);
        break;
    case XAV1_ENC_IOCTL_WAIT_MSG:
        ret = xav1_enc_wait_msg(xav1e, filp, arg);
        break;
    case XAV1_ENC_IOCTL_WHOLE_SOFT_RESET:
        __get_user(slice, (u32 *)arg);
        ret = xav1_whole_subsys_soft_reset(xav1e, slice);
        break;
    case XAV1_ENC_IOCTL_TWO_SLICE_SOFT_RESET:
        ret = xav1_whole_subsys_soft_reset(xav1e, 0);
        ret |= xav1_whole_subsys_soft_reset(xav1e, 1);
        break;
    case XAV1_ENC_IOCTL_BOOT_LOCK_ACQUIRE:
        TRACE("%p WAITING\n", filp);
        ret = down_killable(&xav1e->boot_sem);
        TRACE("%p ACQUIRED\n", filp);
        xav1e->boot_sem_owner = filp;
        break;
    case XAV1_ENC_IOCTL_BOOT_LOCK_RELEASE:
        xav1e->boot_sem_owner = NULL;
        TRACE("%p RELEASING\n", filp);
        up(&xav1e->boot_sem);
        break;
    case PWRTEST_XAV1_RESET_KEEP:
        __get_user(slice, (u32 *)arg);
        sn_pri(tdev, SN_DBG,
            "xav1_enc: %s, slice: %d.\n", __func__, slice);
        sys_xav1_whole_subsys_soft_reset(xav1e->tdev, slice);
        cancel_slot_allocation(xav1e, slice);
        ret = 0;
        break;
    case PWRTEST_XAV1_RESET_RELEASE:
        __get_user(slice, (u32 *)arg);
        sn_pri(tdev, SN_DBG,
            "xav1_enc: %s, slice: %d.\n", __func__, slice);
        sys_xav1_whole_subsys_release_soft_reset(xav1e->tdev, slice);
        enable_mbox15_host_intr(xav1e, slice);
        begin_slot_allocation(xav1e, slice);
        sn_pri(tdev, SN_INF, "xav1_enc: (%s, %d), slice: %d PWRTEST_XAV1_RESET_RELEASE over.\n", __func__, __LINE__, slice);
        ret = 0;
        break;
    default:
        sn_pri(tdev, SN_ERR,
            "xav1_enc: %s, cmd:0x%x error.\n", __func__, cmd);
        ret = -EINVAL;
    }

    switch(cmd)
    { /* After XAV1 reset, load encoder firmwares to FPS sub system */
    case XAV1_ENC_IOCTL_WHOLE_SOFT_RESET:
    case XAV1_ENC_IOCTL_TWO_SLICE_SOFT_RESET:
    case PWRTEST_XAV1_RESET_RELEASE:
        ret = load_fps_subsystem(tdev);
        break;

    default:
        break;
    }

    return ret;
}

void xav1_close(struct sn_tranx_t *tdev, struct file *filp)
{
    struct xav1_enc *xav1e = tdev->modules[SN_MODULE_XAV1_ENC];
    struct slice* slice = xav1e->slices;
    int i;
    TRACE("xav1_close\n");
    for (i = 0; i < MAX_SLICE_NUM; ++i, ++slice) {
        unsigned long flags;
        int j;
        struct slot* slot = slice->slots;
        spin_lock_irqsave(&xav1e->reserve_lock, flags);
        for (j = 0; j < MAX_SLOTS_PER_SLICE; ++j, ++slot) {
            if (slot->filp == filp && !slot->pendingDelete) {
                int retry = 1000;
                slot->pendingDelete = 1;
                {
                    struct xav1_msg msg;
                    msg.slice = i;
                    msg.count = 1;
                    msg.content[0] = 255 | (j << 10);
                    while (retry-- && xav1_enc_send_msg_raw(xav1e, &msg, 0) < 0) { // no lock: We're already holding reserve_lock to exclude the possibility of reset.
                        spin_unlock_irqrestore(&xav1e->reserve_lock, flags);
                        msleep(1);
                        spin_lock_irqsave(&xav1e->reserve_lock, flags);
                    }
                }
                spin_unlock_irqrestore(&xav1e->reserve_lock, flags);
                TRACE("Closing dangling slot %d:%d for %p\n", i, j, filp);
                if (retry < 0) {
                    xav1e->fatal_fw_error = true;
                    sn_pri(tdev, SN_ERR, "Firmware is unresponsive on slice %d\n", slice - xav1e->slices);
                }
                spin_lock_irqsave(&xav1e->reserve_lock, flags);
            }
        }
        spin_unlock_irqrestore(&xav1e->reserve_lock, flags);
    }

    if (filp == xav1e->boot_sem_owner) {
        xav1e->boot_sem_owner = NULL;
        TRACE("%p CLOSE RELEASE\n", filp);
        up(&xav1e->boot_sem);
    }
}
